"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.utils.event_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import hashlib
import os
import shutil
import time

import pytest
import salt.config
import salt.ext.tornado.ioloop
import salt.utils.event
import salt.utils.stringutils
import zmq
import zmq.eventloop.ioloop
from salt.ext.tornado.testing import AsyncTestCase
from saltfactories.utils.processes import terminate_process
from tests.support.events import eventpublisher_process, eventsender_process
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, expectedFailure, skipIf

NO_LONG_IPC = False
if getattr(zmq, "IPC_PATH_MAX_LEN", 103) <= 103:
    NO_LONG_IPC = True


@skipIf(
    NO_LONG_IPC, "This system does not support long IPC paths. Skipping event tests!"
)
class TestSaltEvent(TestCase):
    def setUp(self):
        self.sock_dir = os.path.join(RUNTIME_VARS.TMP, "test-socks")
        if not os.path.exists(self.sock_dir):
            os.makedirs(self.sock_dir)
        self.addCleanup(shutil.rmtree, self.sock_dir, ignore_errors=True)

    def assertGotEvent(self, evt, data, msg=None):
        self.assertIsNotNone(evt, msg)
        for key in data:
            self.assertIn(key, evt, "{}: Key {} missing".format(msg, key))
            assertMsg = "{0}: Key {1} value mismatch, {2} != {3}"
            assertMsg = assertMsg.format(msg, key, data[key], evt[key])
            self.assertEqual(data[key], evt[key], assertMsg)

    def test_master_event(self):
        with salt.utils.event.MasterEvent(self.sock_dir, listen=False) as me:
            self.assertEqual(
                me.puburi,
                "{}".format(os.path.join(self.sock_dir, "master_event_pub.ipc")),
            )
            self.assertEqual(
                me.pulluri,
                "{}".format(os.path.join(self.sock_dir, "master_event_pull.ipc")),
            )

    def test_minion_event(self):
        opts = dict(id="foo", sock_dir=self.sock_dir)
        id_hash = hashlib.sha256(
            salt.utils.stringutils.to_bytes(opts["id"])
        ).hexdigest()[:10]
        with salt.utils.event.MinionEvent(opts, listen=False) as me:
            self.assertEqual(
                me.puburi,
                "{}".format(
                    os.path.join(
                        self.sock_dir, "minion_event_{}_pub.ipc".format(id_hash)
                    )
                ),
            )
            self.assertEqual(
                me.pulluri,
                "{}".format(
                    os.path.join(
                        self.sock_dir, "minion_event_{}_pull.ipc".format(id_hash)
                    )
                ),
            )

    def test_minion_event_tcp_ipc_mode(self):
        opts = dict(id="foo", ipc_mode="tcp")
        with salt.utils.event.MinionEvent(opts, listen=False) as me:
            self.assertEqual(me.puburi, 4510)
            self.assertEqual(me.pulluri, 4511)

    def test_minion_event_no_id(self):
        with salt.utils.event.MinionEvent(
            dict(sock_dir=self.sock_dir), listen=False
        ) as me:
            id_hash = hashlib.sha256(salt.utils.stringutils.to_bytes("")).hexdigest()[
                :10
            ]
            self.assertEqual(
                me.puburi,
                "{}".format(
                    os.path.join(
                        self.sock_dir, "minion_event_{}_pub.ipc".format(id_hash)
                    )
                ),
            )
            self.assertEqual(
                me.pulluri,
                "{}".format(
                    os.path.join(
                        self.sock_dir, "minion_event_{}_pull.ipc".format(id_hash)
                    )
                ),
            )

    @pytest.mark.slow_test
    def test_event_single(self):
        """Test a single event is received"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(tag="evt1")
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_single_no_block(self):
        """Test a single event is received, no block"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                start = time.time()
                finish = start + 5
                evt1 = me.get_event(wait=0, tag="evt1", no_block=True)
                # We should get None and way before the 5 seconds wait since it's
                # non-blocking, otherwise it would wait for an event which we
                # didn't even send
                self.assertIsNone(evt1, None)
                self.assertLess(start, finish)
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(wait=0, tag="evt1")
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_single_wait_0_no_block_False(self):
        """Test a single event is received with wait=0 and no_block=False and doesn't spin the while loop"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                # This is too fast and will be None but assures we're not blocking
                evt1 = me.get_event(wait=0, tag="evt1", no_block=False)
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_timeout(self):
        """Test no event is received if the timeout is reached"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(tag="evt1")
                self.assertGotEvent(evt1, {"data": "foo1"})
                evt2 = me.get_event(tag="evt1")
                self.assertIsNone(evt2)

    @pytest.mark.slow_test
    def test_event_no_timeout(self):
        """Test no wait timeout, we should block forever, until we get one"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                with eventsender_process({"data": "foo2"}, "evt2", self.sock_dir, 5):
                    evt = me.get_event(tag="evt2", wait=0, no_block=False)
                self.assertGotEvent(evt, {"data": "foo2"})

    @pytest.mark.slow_test
    def test_event_matching(self):
        """Test a startswith match"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(tag="ev")
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_matching_regex(self):
        """Test a regex match"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(tag="^ev", match_type="regex")
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_matching_all(self):
        """Test an all match"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event(tag="")
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_matching_all_when_tag_is_None(self):
        """Test event matching all when not passing a tag"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                evt1 = me.get_event()
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_not_subscribed(self):
        """Test get_event drops non-subscribed events"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                me.fire_event({"data": "foo2"}, "evt2")
                evt2 = me.get_event(tag="evt2")
                evt1 = me.get_event(tag="evt1")
                self.assertGotEvent(evt2, {"data": "foo2"})
                self.assertIsNone(evt1)

    @pytest.mark.slow_test
    def test_event_subscription_cache(self):
        """Test subscriptions cache a message until requested"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.subscribe("evt1")
                me.fire_event({"data": "foo1"}, "evt1")
                me.fire_event({"data": "foo2"}, "evt2")
                evt2 = me.get_event(tag="evt2")
                evt1 = me.get_event(tag="evt1")
                self.assertGotEvent(evt2, {"data": "foo2"})
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_subscriptions_cache_regex(self):
        """Test regex subscriptions cache a message until requested"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.subscribe("e..1$", "regex")
                me.fire_event({"data": "foo1"}, "evt1")
                me.fire_event({"data": "foo2"}, "evt2")
                evt2 = me.get_event(tag="evt2")
                evt1 = me.get_event(tag="evt1")
                self.assertGotEvent(evt2, {"data": "foo2"})
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_multiple_clients(self):
        """Test event is received by multiple clients"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(
                self.sock_dir, listen=True
            ) as me1, salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me2:
                # We need to sleep here to avoid a race condition wherein
                # the second socket may not be connected by the time the first socket
                # sends the event.
                time.sleep(0.5)
                me1.fire_event({"data": "foo1"}, "evt1")
                evt1 = me1.get_event(tag="evt1")
                self.assertGotEvent(evt1, {"data": "foo1"})
                evt2 = me2.get_event(tag="evt1")
                self.assertGotEvent(evt2, {"data": "foo1"})

    @expectedFailure
    def test_event_nested_sub_all(self):
        """Test nested event subscriptions do not drop events, get event for all tags"""
        # Show why not to call get_event(tag='')
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                me.fire_event({"data": "foo1"}, "evt1")
                me.fire_event({"data": "foo2"}, "evt2")
                evt2 = me.get_event(tag="")
                evt1 = me.get_event(tag="")
                self.assertGotEvent(evt2, {"data": "foo2"})
                self.assertGotEvent(evt1, {"data": "foo1"})

    @pytest.mark.slow_test
    def test_event_many(self):
        """Test a large number of events, one at a time"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                for i in range(500):
                    me.fire_event({"data": "{}".format(i)}, "testevents")
                    evt = me.get_event(tag="testevents")
                    self.assertGotEvent(
                        evt, {"data": "{}".format(i)}, "Event {}".format(i)
                    )

    @pytest.mark.slow_test
    def test_event_many_backlog(self):
        """Test a large number of events, send all then recv all"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                # Must not exceed zmq HWM
                for i in range(500):
                    me.fire_event({"data": "{}".format(i)}, "testevents")
                for i in range(500):
                    evt = me.get_event(tag="testevents")
                    self.assertGotEvent(
                        evt, {"data": "{}".format(i)}, "Event {}".format(i)
                    )

    # Test the fire_master function. As it wraps the underlying fire_event,
    # we don't need to perform extensive testing.
    @pytest.mark.slow_test
    def test_send_master_event(self):
        """Tests that sending an event through fire_master generates expected event"""
        with eventpublisher_process(self.sock_dir):
            with salt.utils.event.MasterEvent(self.sock_dir, listen=True) as me:
                data = {"data": "foo1"}
                me.fire_master(data, "test_master")

                evt = me.get_event(tag="fire_master")
                self.assertGotEvent(
                    evt,
                    {
                        "data": data,
                        "tag": "test_master",
                        "events": None,
                        "pretag": None,
                    },
                )


class TestAsyncEventPublisher(AsyncTestCase):
    def get_new_ioloop(self):
        return salt.ext.tornado.ioloop.IOLoop()

    def setUp(self):
        super().setUp()
        self.sock_dir = os.path.join(RUNTIME_VARS.TMP, "test-socks")
        if not os.path.exists(self.sock_dir):
            os.makedirs(self.sock_dir)
        self.addCleanup(shutil.rmtree, self.sock_dir, ignore_errors=True)
        self.opts = {"sock_dir": self.sock_dir}
        self.publisher = salt.utils.event.AsyncEventPublisher(
            self.opts,
            self.io_loop,
        )
        self.event = salt.utils.event.get_event(
            "minion", opts=self.opts, io_loop=self.io_loop
        )
        self.event.subscribe("")
        self.event.set_event_handler(self._handle_publish)

    def stop(self, _arg=None, **kwargs):
        self.publisher.close()
        self.event.destroy()
        super().stop(_arg=_arg, **kwargs)

    def _handle_publish(self, raw):
        self.tag, self.data = salt.utils.event.SaltEvent.unpack(raw)
        self.stop()

    def test_event_subscription(self):
        """Test a single event is received"""
        with salt.utils.event.MinionEvent(self.opts, listen=True) as me:
            me.fire_event({"data": "foo1"}, "evt1")
            self.wait()
            evt1 = me.get_event(tag="evt1")
            self.assertEqual(self.tag, "evt1")
            self.data.pop("_stamp")  # drop the stamp
            self.assertEqual(self.data, {"data": "foo1"})

    def test_event_unsubscribe_remove_error(self):
        with salt.utils.event.MinionEvent(self.opts, listen=True) as me:
            tag = "evt1"
            me.fire_event({"data": "foo1"}, tag)

            # Make sure no remove error is raised when tag is not found
            for _ in range(2):
                me.unsubscribe(tag)

            me.unsubscribe("tag_does_not_exist")


class TestEventReturn(TestCase):
    @pytest.mark.slow_test
    def test_event_return(self):
        evt = None
        try:
            evt = salt.utils.event.EventReturn(salt.config.DEFAULT_MASTER_OPTS.copy())
            evt.start()
        except TypeError as exc:
            if "object" in str(exc):
                self.fail("'{}' TypeError should have not been raised".format(exc))
        finally:
            if evt is not None:
                terminate_process(evt.pid, kill_children=True)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.zk_concurrency as zk_concurrency
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {zk_concurrency: {}}


def test_lock():
    """
    Test to block state execution until you are able to get the lock
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    with patch.dict(zk_concurrency.__opts__, {"test": True}):
        ret.update({"comment": "Attempt to acquire lock", "result": None})
        assert zk_concurrency.lock("salt", "dude") == ret

    with patch.dict(zk_concurrency.__opts__, {"test": False}):
        mock = MagicMock(return_value=True)
        with patch.dict(zk_concurrency.__salt__, {"zk_concurrency.lock": mock}):
            ret.update({"comment": "lock acquired", "result": True})
            assert zk_concurrency.lock("salt", "dude", "stack") == ret


def test_unlock():
    """
    Test to remove lease from semaphore
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    with patch.dict(zk_concurrency.__opts__, {"test": True}):
        ret.update({"comment": "Released lock if it is here", "result": None})
        assert zk_concurrency.unlock("salt") == ret

    with patch.dict(zk_concurrency.__opts__, {"test": False}):
        mock = MagicMock(return_value=True)
        with patch.dict(zk_concurrency.__salt__, {"zk_concurrency.unlock": mock}):
            ret.update({"comment": "", "result": True})
            assert zk_concurrency.unlock("salt", identifier="stack") == ret


def test_min_party():
    """
    Test to ensure min party of nodes and the blocking behavior
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    with patch.dict(zk_concurrency.__opts__, {"test": True}):
        ret.update({"comment": "Attempt to ensure min_party", "result": None})
        assert zk_concurrency.min_party("salt", "dude", 1) == ret

    with patch.dict(zk_concurrency.__opts__, {"test": False}):
        mock = MagicMock(return_value=["1", "2", "3"])
        with patch.dict(
            zk_concurrency.__salt__, {"zk_concurrency.party_members": mock}
        ):
            ret.update({"comment": "Currently 3 nodes, which is >= 2", "result": True})
            assert zk_concurrency.min_party("salt", "dude", 2) == ret
            ret.update(
                {
                    "comment": "Blocked until 2 nodes were available. "
                    + "Unblocked after 3 nodes became available",
                    "result": True,
                }
            )
            assert zk_concurrency.min_party("salt", "dude", 2, True) == ret
            ret.update({"comment": "Currently 3 nodes, which is < 4", "result": False})
            assert zk_concurrency.min_party("salt", "dude", 4) == ret



# pylint: skip-file
from __future__ import absolute_import, division, print_function
from salt.ext.tornado.concurrent import Future
from salt.ext.tornado import gen
from salt.ext.tornado import netutil
from salt.ext.tornado.iostream import IOStream, SSLIOStream, PipeIOStream, StreamClosedError
from salt.ext.tornado.httputil import HTTPHeaders
from salt.ext.tornado.log import gen_log, app_log
from salt.ext.tornado.netutil import ssl_wrap_socket
from salt.ext.tornado.stack_context import NullContext
from salt.ext.tornado.tcpserver import TCPServer
from salt.ext.tornado.testing import AsyncHTTPTestCase, AsyncHTTPSTestCase, AsyncTestCase, bind_unused_port, ExpectLog, gen_test
from salt.ext.tornado.test.util import unittest, skipIfNonUnix, refusing_port, skipPypy3V58
from salt.ext.tornado.web import RequestHandler, Application
import errno
import logging
import os
import platform
import socket
import ssl
import sys

try:
    from unittest import mock  # type: ignore
except ImportError:
    try:
        import mock  # type: ignore
    except ImportError:
        mock = None


def _server_ssl_options():
    return dict(
        certfile=os.path.join(os.path.dirname(__file__), 'test.crt'),
        keyfile=os.path.join(os.path.dirname(__file__), 'test.key'),
    )


class HelloHandler(RequestHandler):
    def get(self):
        self.write("Hello")


class TestIOStreamWebMixin(object):
    def _make_client_iostream(self):
        raise NotImplementedError()

    def get_app(self):
        return Application([('/', HelloHandler)])

    def test_connection_closed(self):
        # When a server sends a response and then closes the connection,
        # the client must be allowed to read the data before the IOStream
        # closes itself.  Epoll reports closed connections with a separate
        # EPOLLRDHUP event delivered at the same time as the read event,
        # while kqueue reports them as a second read/write event with an EOF
        # flag.
        response = self.fetch("/", headers={"Connection": "close"})
        response.rethrow()

    def test_read_until_close(self):
        stream = self._make_client_iostream()
        stream.connect(('127.0.0.1', self.get_http_port()), callback=self.stop)
        self.wait()
        stream.write(b"GET / HTTP/1.0\r\n\r\n")

        stream.read_until_close(self.stop)
        data = self.wait()
        self.assertTrue(data.startswith(b"HTTP/1.1 200"))
        self.assertTrue(data.endswith(b"Hello"))

    def test_read_zero_bytes(self):
        self.stream = self._make_client_iostream()
        self.stream.connect(("127.0.0.1", self.get_http_port()),
                            callback=self.stop)
        self.wait()
        self.stream.write(b"GET / HTTP/1.0\r\n\r\n")

        # normal read
        self.stream.read_bytes(9, self.stop)
        data = self.wait()
        self.assertEqual(data, b"HTTP/1.1 ")

        # zero bytes
        self.stream.read_bytes(0, self.stop)
        data = self.wait()
        self.assertEqual(data, b"")

        # another normal read
        self.stream.read_bytes(3, self.stop)
        data = self.wait()
        self.assertEqual(data, b"200")

        self.stream.close()

    def test_write_while_connecting(self):
        stream = self._make_client_iostream()
        connected = [False]

        def connected_callback():
            connected[0] = True
            self.stop()
        stream.connect(("127.0.0.1", self.get_http_port()),
                       callback=connected_callback)
        # unlike the previous tests, try to write before the connection
        # is complete.
        written = [False]

        def write_callback():
            written[0] = True
            self.stop()
        stream.write(b"GET / HTTP/1.0\r\nConnection: close\r\n\r\n",
                     callback=write_callback)
        self.assertTrue(not connected[0])
        # by the time the write has flushed, the connection callback has
        # also run
        try:
            self.wait(lambda: connected[0] and written[0])
        finally:
            logging.debug((connected, written))

        stream.read_until_close(self.stop)
        data = self.wait()
        self.assertTrue(data.endswith(b"Hello"))

        stream.close()

    @gen_test
    def test_future_interface(self):
        """Basic test of IOStream's ability to return Futures."""
        stream = self._make_client_iostream()
        connect_result = yield stream.connect(
            ("127.0.0.1", self.get_http_port()))
        self.assertIs(connect_result, stream)
        yield stream.write(b"GET / HTTP/1.0\r\n\r\n")
        first_line = yield stream.read_until(b"\r\n")
        self.assertEqual(first_line, b"HTTP/1.1 200 OK\r\n")
        # callback=None is equivalent to no callback.
        header_data = yield stream.read_until(b"\r\n\r\n", callback=None)
        headers = HTTPHeaders.parse(header_data.decode('latin1'))
        content_length = int(headers['Content-Length'])
        body = yield stream.read_bytes(content_length)
        self.assertEqual(body, b'Hello')
        stream.close()

    @gen_test
    def test_future_close_while_reading(self):
        stream = self._make_client_iostream()
        yield stream.connect(("127.0.0.1", self.get_http_port()))
        yield stream.write(b"GET / HTTP/1.0\r\n\r\n")
        with self.assertRaises(StreamClosedError):
            yield stream.read_bytes(1024 * 1024)
        stream.close()

    @gen_test
    def test_future_read_until_close(self):
        # Ensure that the data comes through before the StreamClosedError.
        stream = self._make_client_iostream()
        yield stream.connect(("127.0.0.1", self.get_http_port()))
        yield stream.write(b"GET / HTTP/1.0\r\nConnection: close\r\n\r\n")
        yield stream.read_until(b"\r\n\r\n")
        body = yield stream.read_until_close()
        self.assertEqual(body, b"Hello")

        # Nothing else to read; the error comes immediately without waiting
        # for yield.
        with self.assertRaises(StreamClosedError):
            stream.read_bytes(1)


class TestIOStreamMixin(object):
    def _make_server_iostream(self, connection, **kwargs):
        raise NotImplementedError()

    def _make_client_iostream(self, connection, **kwargs):
        raise NotImplementedError()

    def make_iostream_pair(self, **kwargs):
        listener, port = bind_unused_port()
        streams = [None, None]

        def accept_callback(connection, address):
            streams[0] = self._make_server_iostream(connection, **kwargs)
            self.stop()

        def connect_callback():
            streams[1] = client_stream
            self.stop()
        netutil.add_accept_handler(listener, accept_callback,
                                   io_loop=self.io_loop)
        client_stream = self._make_client_iostream(socket.socket(), **kwargs)
        client_stream.connect(('127.0.0.1', port),
                              callback=connect_callback)
        self.wait(condition=lambda: all(streams))
        self.io_loop.remove_handler(listener.fileno())
        listener.close()
        return streams

    def test_streaming_callback_with_data_in_buffer(self):
        server, client = self.make_iostream_pair()
        client.write(b"abcd\r\nefgh")
        server.read_until(b"\r\n", self.stop)
        data = self.wait()
        self.assertEqual(data, b"abcd\r\n")

        def closed_callback(chunk):
            self.fail()
        server.read_until_close(callback=closed_callback,
                                streaming_callback=self.stop)
        # self.io_loop.add_timeout(self.io_loop.time() + 0.01, self.stop)
        data = self.wait()
        self.assertEqual(data, b"efgh")
        server.close()
        client.close()

    def test_write_zero_bytes(self):
        # Attempting to write zero bytes should run the callback without
        # going into an infinite loop.
        server, client = self.make_iostream_pair()
        server.write(b'', callback=self.stop)
        self.wait()
        server.close()
        client.close()

    def test_connection_refused(self):
        # When a connection is refused, the connect callback should not
        # be run.  (The kqueue IOLoop used to behave differently from the
        # epoll IOLoop in this respect)
        cleanup_func, port = refusing_port()
        self.addCleanup(cleanup_func)
        stream = IOStream(socket.socket(), self.io_loop)
        self.connect_called = False

        def connect_callback():
            self.connect_called = True
            self.stop()
        stream.set_close_callback(self.stop)
        # log messages vary by platform and ioloop implementation
        with ExpectLog(gen_log, ".*", required=False):
            stream.connect(("127.0.0.1", port), connect_callback)
            self.wait()
        self.assertFalse(self.connect_called)
        self.assertTrue(isinstance(stream.error, socket.error), stream.error)
        if sys.platform != 'cygwin':
            _ERRNO_CONNREFUSED = (errno.ECONNREFUSED,)
            if hasattr(errno, "WSAECONNREFUSED"):
                _ERRNO_CONNREFUSED += (errno.WSAECONNREFUSED,)
            # cygwin's errnos don't match those used on native windows python
            self.assertTrue(stream.error.args[0] in _ERRNO_CONNREFUSED)

    @unittest.skipIf(mock is None, 'mock package not present')
    def test_gaierror(self):
        # Test that IOStream sets its exc_info on getaddrinfo error.
        # It's difficult to reliably trigger a getaddrinfo error;
        # some resolvers own't even return errors for malformed names,
        # so we mock it instead. If IOStream changes to call a Resolver
        # before sock.connect, the mock target will need to change too.
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        stream = IOStream(s, io_loop=self.io_loop)
        stream.set_close_callback(self.stop)
        with mock.patch('socket.socket.connect',
                        side_effect=socket.gaierror(errno.EIO, 'boom')):
            with ExpectLog(gen_log, "Connect error"):
                stream.connect(('localhost', 80), callback=self.stop)
                self.wait()
                self.assertIsInstance(stream.error, socket.gaierror)

    def test_read_callback_error(self):
        # Test that IOStream sets its exc_info when a read callback throws
        server, client = self.make_iostream_pair()
        try:
            server.set_close_callback(self.stop)
            with ExpectLog(
                app_log, "(Uncaught exception|Exception in callback)"
            ):
                # Clear ExceptionStackContext so IOStream catches error
                with NullContext():
                    server.read_bytes(1, callback=lambda data: 1 / 0)
                client.write(b"1")
                self.wait()
            self.assertTrue(isinstance(server.error, ZeroDivisionError))
        finally:
            server.close()
            client.close()

    def test_streaming_callback(self):
        server, client = self.make_iostream_pair()
        try:
            chunks = []
            final_called = []

            def streaming_callback(data):
                chunks.append(data)
                self.stop()

            def final_callback(data):
                self.assertFalse(data)
                final_called.append(True)
                self.stop()
            server.read_bytes(6, callback=final_callback,
                              streaming_callback=streaming_callback)
            client.write(b"1234")
            self.wait(condition=lambda: chunks)
            client.write(b"5678")
            self.wait(condition=lambda: final_called)
            self.assertEqual(chunks, [b"1234", b"56"])

            # the rest of the last chunk is still in the buffer
            server.read_bytes(2, callback=self.stop)
            data = self.wait()
            self.assertEqual(data, b"78")
        finally:
            server.close()
            client.close()

    def test_streaming_until_close(self):
        server, client = self.make_iostream_pair()
        try:
            chunks = []
            closed = [False]

            def streaming_callback(data):
                chunks.append(data)
                self.stop()

            def close_callback(data):
                assert not data, data
                closed[0] = True
                self.stop()
            client.read_until_close(callback=close_callback,
                                    streaming_callback=streaming_callback)
            server.write(b"1234")
            self.wait(condition=lambda: len(chunks) == 1)
            server.write(b"5678", self.stop)
            self.wait()
            server.close()
            self.wait(condition=lambda: closed[0])
            self.assertEqual(chunks, [b"1234", b"5678"])
        finally:
            server.close()
            client.close()

    def test_streaming_until_close_future(self):
        server, client = self.make_iostream_pair()
        try:
            chunks = []

            @gen.coroutine
            def client_task():
                yield client.read_until_close(streaming_callback=chunks.append)

            @gen.coroutine
            def server_task():
                yield server.write(b"1234")
                yield gen.sleep(0.01)
                yield server.write(b"5678")
                server.close()

            @gen.coroutine
            def f():
                yield [client_task(), server_task()]
            self.io_loop.run_sync(f)
            self.assertEqual(chunks, [b"1234", b"5678"])
        finally:
            server.close()
            client.close()

    def test_delayed_close_callback(self):
        # The scenario:  Server closes the connection while there is a pending
        # read that can be served out of buffered data.  The client does not
        # run the close_callback as soon as it detects the close, but rather
        # defers it until after the buffered read has finished.
        server, client = self.make_iostream_pair()
        try:
            client.set_close_callback(self.stop)
            server.write(b"12")
            chunks = []

            def callback1(data):
                chunks.append(data)
                client.read_bytes(1, callback2)
                server.close()

            def callback2(data):
                chunks.append(data)
            client.read_bytes(1, callback1)
            self.wait()  # stopped by close_callback
            self.assertEqual(chunks, [b"1", b"2"])
        finally:
            server.close()
            client.close()

    def test_future_delayed_close_callback(self):
        # Same as test_delayed_close_callback, but with the future interface.
        server, client = self.make_iostream_pair()

        # We can't call make_iostream_pair inside a gen_test function
        # because the ioloop is not reentrant.
        @gen_test
        def f(self):
            server.write(b"12")
            chunks = []
            chunks.append((yield client.read_bytes(1)))
            server.close()
            chunks.append((yield client.read_bytes(1)))
            self.assertEqual(chunks, [b"1", b"2"])
        try:
            f(self)
        finally:
            server.close()
            client.close()

    def test_close_buffered_data(self):
        # Similar to the previous test, but with data stored in the OS's
        # socket buffers instead of the IOStream's read buffer.  Out-of-band
        # close notifications must be delayed until all data has been
        # drained into the IOStream buffer. (epoll used to use out-of-band
        # close events with EPOLLRDHUP, but no longer)
        #
        # This depends on the read_chunk_size being smaller than the
        # OS socket buffer, so make it small.
        server, client = self.make_iostream_pair(read_chunk_size=256)
        try:
            server.write(b"A" * 512)
            client.read_bytes(256, self.stop)
            data = self.wait()
            self.assertEqual(b"A" * 256, data)
            server.close()
            # Allow the close to propagate to the client side of the
            # connection.  Using add_callback instead of add_timeout
            # doesn't seem to work, even with multiple iterations
            self.io_loop.add_timeout(self.io_loop.time() + 0.01, self.stop)
            self.wait()
            client.read_bytes(256, self.stop)
            data = self.wait()
            self.assertEqual(b"A" * 256, data)
        finally:
            server.close()
            client.close()

    def test_read_until_close_after_close(self):
        # Similar to test_delayed_close_callback, but read_until_close takes
        # a separate code path so test it separately.
        server, client = self.make_iostream_pair()
        try:
            server.write(b"1234")
            server.close()
            # Read one byte to make sure the client has received the data.
            # It won't run the close callback as long as there is more buffered
            # data that could satisfy a later read.
            client.read_bytes(1, self.stop)
            data = self.wait()
            self.assertEqual(data, b"1")
            client.read_until_close(self.stop)
            data = self.wait()
            self.assertEqual(data, b"234")
        finally:
            server.close()
            client.close()

    @unittest.skipIf(mock is None, 'mock package not present')
    def test_read_until_close_with_error(self):
        server, client = self.make_iostream_pair()
        try:
            with mock.patch('tornado.iostream.BaseIOStream._try_inline_read',
                            side_effect=IOError('boom')):
                with self.assertRaisesRegexp(IOError, 'boom'):
                    client.read_until_close(self.stop)
        finally:
            server.close()
            client.close()

    def test_streaming_read_until_close_after_close(self):
        # Same as the preceding test but with a streaming_callback.
        # All data should go through the streaming callback,
        # and the final read callback just gets an empty string.
        server, client = self.make_iostream_pair()
        try:
            server.write(b"1234")
            server.close()
            client.read_bytes(1, self.stop)
            data = self.wait()
            self.assertEqual(data, b"1")
            streaming_data = []
            client.read_until_close(self.stop,
                                    streaming_callback=streaming_data.append)
            data = self.wait()
            self.assertEqual(b'', data)
            self.assertEqual(b''.join(streaming_data), b"234")
        finally:
            server.close()
            client.close()

    def test_large_read_until(self):
        # Performance test: read_until used to have a quadratic component
        # so a read_until of 4MB would take 8 seconds; now it takes 0.25
        # seconds.
        server, client = self.make_iostream_pair()
        try:
            # This test fails on pypy with ssl.  I think it's because
            # pypy's gc defeats moves objects, breaking the
            # "frozen write buffer" assumption.
            if (isinstance(server, SSLIOStream) and
                    platform.python_implementation() == 'PyPy'):
                raise unittest.SkipTest(
                    "pypy gc causes problems with openssl")
            NUM_KB = 4096
            for i in range(NUM_KB):
                client.write(b"A" * 1024)
            client.write(b"\r\n")
            server.read_until(b"\r\n", self.stop)
            data = self.wait()
            self.assertEqual(len(data), NUM_KB * 1024 + 2)
        finally:
            server.close()
            client.close()

    def test_close_callback_with_pending_read(self):
        # Regression test for a bug that was introduced in 2.3
        # where the IOStream._close_callback would never be called
        # if there were pending reads.
        OK = b"OK\r\n"
        server, client = self.make_iostream_pair()
        client.set_close_callback(self.stop)
        try:
            server.write(OK)
            client.read_until(b"\r\n", self.stop)
            res = self.wait()
            self.assertEqual(res, OK)

            server.close()
            client.read_until(b"\r\n", lambda x: x)
            # If _close_callback (self.stop) is not called,
            # an AssertionError: Async operation timed out after 5 seconds
            # will be raised.
            res = self.wait()
            self.assertTrue(res is None)
        finally:
            server.close()
            client.close()

    @skipIfNonUnix
    @skipPypy3V58
    def test_inline_read_error(self):
        # An error on an inline read is raised without logging (on the
        # assumption that it will eventually be noticed or logged further
        # up the stack).
        #
        # This test is posix-only because windows os.close() doesn't work
        # on socket FDs, but we can't close the socket object normally
        # because we won't get the error we want if the socket knows
        # it's closed.
        server, client = self.make_iostream_pair()
        try:
            os.close(server.socket.fileno())
            with self.assertRaises(socket.error):
                server.read_bytes(1, lambda data: None)
        finally:
            server.close()
            client.close()

    @skipPypy3V58
    def test_async_read_error_logging(self):
        # Socket errors on asynchronous reads should be logged (but only
        # once).
        server, client = self.make_iostream_pair()
        server.set_close_callback(self.stop)
        try:
            # Start a read that will be fulfilled asynchronously.
            server.read_bytes(1, lambda data: None)
            client.write(b'a')
            # Stub out read_from_fd to make it fail.

            def fake_read_from_fd():
                os.close(server.socket.fileno())
                server.__class__.read_from_fd(server)
            server.read_from_fd = fake_read_from_fd
            # This log message is from _handle_read (not read_from_fd).
            with ExpectLog(gen_log, "error on read"):
                self.wait()
        finally:
            server.close()
            client.close()

    def test_future_close_callback(self):
        # Regression test for interaction between the Future read interfaces
        # and IOStream._maybe_add_error_listener.
        server, client = self.make_iostream_pair()
        closed = [False]

        def close_callback():
            closed[0] = True
            self.stop()
        server.set_close_callback(close_callback)
        try:
            client.write(b'a')
            future = server.read_bytes(1)
            self.io_loop.add_future(future, self.stop)
            self.assertEqual(self.wait().result(), b'a')
            self.assertFalse(closed[0])
            client.close()
            self.wait()
            self.assertTrue(closed[0])
        finally:
            server.close()
            client.close()

    def test_write_memoryview(self):
        server, client = self.make_iostream_pair()
        try:
            client.read_bytes(4, self.stop)
            server.write(memoryview(b"hello"))
            data = self.wait()
            self.assertEqual(data, b"hell")
        finally:
            server.close()
            client.close()

    def test_read_bytes_partial(self):
        server, client = self.make_iostream_pair()
        try:
            # Ask for more than is available with partial=True
            client.read_bytes(50, self.stop, partial=True)
            server.write(b"hello")
            data = self.wait()
            self.assertEqual(data, b"hello")

            # Ask for less than what is available; num_bytes is still
            # respected.
            client.read_bytes(3, self.stop, partial=True)
            server.write(b"world")
            data = self.wait()
            self.assertEqual(data, b"wor")

            # Partial reads won't return an empty string, but read_bytes(0)
            # will.
            client.read_bytes(0, self.stop, partial=True)
            data = self.wait()
            self.assertEqual(data, b'')
        finally:
            server.close()
            client.close()

    def test_read_until_max_bytes(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Extra room under the limit
            client.read_until(b"def", self.stop, max_bytes=50)
            server.write(b"abcdef")
            data = self.wait()
            self.assertEqual(data, b"abcdef")

            # Just enough space
            client.read_until(b"def", self.stop, max_bytes=6)
            server.write(b"abcdef")
            data = self.wait()
            self.assertEqual(data, b"abcdef")

            # Not enough space, but we don't know it until all we can do is
            # log a warning and close the connection.
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until(b"def", self.stop, max_bytes=5)
                server.write(b"123456")
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_read_until_max_bytes_inline(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Similar to the error case in the previous test, but the
            # server writes first so client reads are satisfied
            # inline.  For consistency with the out-of-line case, we
            # do not raise the error synchronously.
            server.write(b"123456")
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until(b"def", self.stop, max_bytes=5)
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_read_until_max_bytes_ignores_extra(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Even though data that matches arrives the same packet that
            # puts us over the limit, we fail the request because it was not
            # found within the limit.
            server.write(b"abcdef")
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until(b"def", self.stop, max_bytes=5)
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_read_until_regex_max_bytes(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Extra room under the limit
            client.read_until_regex(b"def", self.stop, max_bytes=50)
            server.write(b"abcdef")
            data = self.wait()
            self.assertEqual(data, b"abcdef")

            # Just enough space
            client.read_until_regex(b"def", self.stop, max_bytes=6)
            server.write(b"abcdef")
            data = self.wait()
            self.assertEqual(data, b"abcdef")

            # Not enough space, but we don't know it until all we can do is
            # log a warning and close the connection.
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until_regex(b"def", self.stop, max_bytes=5)
                server.write(b"123456")
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_read_until_regex_max_bytes_inline(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Similar to the error case in the previous test, but the
            # server writes first so client reads are satisfied
            # inline.  For consistency with the out-of-line case, we
            # do not raise the error synchronously.
            server.write(b"123456")
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until_regex(b"def", self.stop, max_bytes=5)
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_read_until_regex_max_bytes_ignores_extra(self):
        server, client = self.make_iostream_pair()
        client.set_close_callback(lambda: self.stop("closed"))
        try:
            # Even though data that matches arrives the same packet that
            # puts us over the limit, we fail the request because it was not
            # found within the limit.
            server.write(b"abcdef")
            with ExpectLog(gen_log, "Unsatisfiable read"):
                client.read_until_regex(b"def", self.stop, max_bytes=5)
                data = self.wait()
            self.assertEqual(data, "closed")
        finally:
            server.close()
            client.close()

    def test_small_reads_from_large_buffer(self):
        # 10KB buffer size, 100KB available to read.
        # Read 1KB at a time and make sure that the buffer is not eagerly
        # filled.
        server, client = self.make_iostream_pair(max_buffer_size=10 * 1024)
        try:
            server.write(b"a" * 1024 * 100)
            for i in range(100):
                client.read_bytes(1024, self.stop)
                data = self.wait()
                self.assertEqual(data, b"a" * 1024)
        finally:
            server.close()
            client.close()

    def test_small_read_untils_from_large_buffer(self):
        # 10KB buffer size, 100KB available to read.
        # Read 1KB at a time and make sure that the buffer is not eagerly
        # filled.
        server, client = self.make_iostream_pair(max_buffer_size=10 * 1024)
        try:
            server.write((b"a" * 1023 + b"\n") * 100)
            for i in range(100):
                client.read_until(b"\n", self.stop, max_bytes=4096)
                data = self.wait()
                self.assertEqual(data, b"a" * 1023 + b"\n")
        finally:
            server.close()
            client.close()

    def test_flow_control(self):
        MB = 1024 * 1024
        server, client = self.make_iostream_pair(max_buffer_size=5 * MB)
        try:
            # Client writes more than the server will accept.
            client.write(b"a" * 10 * MB)
            # The server pauses while reading.
            server.read_bytes(MB, self.stop)
            self.wait()
            self.io_loop.call_later(0.1, self.stop)
            self.wait()
            # The client's writes have been blocked; the server can
            # continue to read gradually.
            for i in range(9):
                server.read_bytes(MB, self.stop)
                self.wait()
        finally:
            server.close()
            client.close()

    def test_future_write(self):
        """
        Test that write() Futures are never orphaned.
        """
        # Run concurrent writers that will write enough bytes so as to
        # clog the socket buffer and accumulate bytes in our write buffer.
        m, n = 10000, 1000
        nproducers = 10
        total_bytes = m * n * nproducers
        server, client = self.make_iostream_pair(max_buffer_size=total_bytes)

        @gen.coroutine
        def produce():
            data = b'x' * m
            for i in range(n):
                yield server.write(data)

        @gen.coroutine
        def consume():
            nread = 0
            while nread < total_bytes:
                res = yield client.read_bytes(m)
                nread += len(res)

        @gen.coroutine
        def main():
            yield [produce() for i in range(nproducers)] + [consume()]

        try:
            self.io_loop.run_sync(main)
        finally:
            server.close()
            client.close()


class TestIOStreamWebHTTP(TestIOStreamWebMixin, AsyncHTTPTestCase):
    def _make_client_iostream(self):
        return IOStream(socket.socket(), io_loop=self.io_loop)


class TestIOStreamWebHTTPS(TestIOStreamWebMixin, AsyncHTTPSTestCase):
    def _make_client_iostream(self):
        return SSLIOStream(socket.socket(), io_loop=self.io_loop,
                           ssl_options=dict(cert_reqs=ssl.CERT_NONE))


class TestIOStream(TestIOStreamMixin, AsyncTestCase):
    def _make_server_iostream(self, connection, **kwargs):
        return IOStream(connection, **kwargs)

    def _make_client_iostream(self, connection, **kwargs):
        return IOStream(connection, **kwargs)


class TestIOStreamSSL(TestIOStreamMixin, AsyncTestCase):
    def _make_server_iostream(self, connection, **kwargs):
        connection = ssl.wrap_socket(connection,
                                     server_side=True,
                                     do_handshake_on_connect=False,
                                     **_server_ssl_options())
        return SSLIOStream(connection, io_loop=self.io_loop, **kwargs)

    def _make_client_iostream(self, connection, **kwargs):
        return SSLIOStream(connection, io_loop=self.io_loop,
                           ssl_options=dict(cert_reqs=ssl.CERT_NONE),
                           **kwargs)


# This will run some tests that are basically redundant but it's the
# simplest way to make sure that it works to pass an SSLContext
# instead of an ssl_options dict to the SSLIOStream constructor.
@unittest.skipIf(not hasattr(ssl, 'SSLContext'), 'ssl.SSLContext not present')
class TestIOStreamSSLContext(TestIOStreamMixin, AsyncTestCase):
    def _make_server_iostream(self, connection, **kwargs):
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        context.load_cert_chain(
            os.path.join(os.path.dirname(__file__), 'test.crt'),
            os.path.join(os.path.dirname(__file__), 'test.key'))
        connection = ssl_wrap_socket(connection, context,
                                     server_side=True,
                                     do_handshake_on_connect=False)
        return SSLIOStream(connection, io_loop=self.io_loop, **kwargs)

    def _make_client_iostream(self, connection, **kwargs):
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        return SSLIOStream(connection, io_loop=self.io_loop,
                           ssl_options=context, **kwargs)


class TestIOStreamStartTLS(AsyncTestCase):
    def setUp(self):
        try:
            super(TestIOStreamStartTLS, self).setUp()
            self.listener, self.port = bind_unused_port()
            self.server_stream = None
            self.server_accepted = Future()
            netutil.add_accept_handler(self.listener, self.accept)
            self.client_stream = IOStream(socket.socket())
            self.io_loop.add_future(self.client_stream.connect(
                ('127.0.0.1', self.port)), self.stop)
            self.wait()
            self.io_loop.add_future(self.server_accepted, self.stop)
            self.wait()
        except Exception as e:
            print(e)
            raise

    def tearDown(self):
        if self.server_stream is not None:
            self.server_stream.close()
        if self.client_stream is not None:
            self.client_stream.close()
        self.listener.close()
        super(TestIOStreamStartTLS, self).tearDown()

    def accept(self, connection, address):
        if self.server_stream is not None:
            self.fail("should only get one connection")
        self.server_stream = IOStream(connection)
        self.server_accepted.set_result(None)

    @gen.coroutine
    def client_send_line(self, line):
        self.client_stream.write(line)
        recv_line = yield self.server_stream.read_until(b"\r\n")
        self.assertEqual(line, recv_line)

    @gen.coroutine
    def server_send_line(self, line):
        self.server_stream.write(line)
        recv_line = yield self.client_stream.read_until(b"\r\n")
        self.assertEqual(line, recv_line)

    def client_start_tls(self, ssl_options=None, server_hostname=None):
        client_stream = self.client_stream
        self.client_stream = None
        return client_stream.start_tls(False, ssl_options, server_hostname)

    def server_start_tls(self, ssl_options=None):
        server_stream = self.server_stream
        self.server_stream = None
        return server_stream.start_tls(True, ssl_options)

    @gen_test
    def test_start_tls_smtp(self):
        # This flow is simplified from RFC 3207 section 5.
        # We don't really need all of this, but it helps to make sure
        # that after realistic back-and-forth traffic the buffers end up
        # in a sane state.
        yield self.server_send_line(b"220 mail.example.com ready\r\n")
        yield self.client_send_line(b"EHLO mail.example.com\r\n")
        yield self.server_send_line(b"250-mail.example.com welcome\r\n")
        yield self.server_send_line(b"250 STARTTLS\r\n")
        yield self.client_send_line(b"STARTTLS\r\n")
        yield self.server_send_line(b"220 Go ahead\r\n")
        client_future = self.client_start_tls(dict(cert_reqs=ssl.CERT_NONE))
        server_future = self.server_start_tls(_server_ssl_options())
        self.client_stream = yield client_future
        self.server_stream = yield server_future
        self.assertTrue(isinstance(self.client_stream, SSLIOStream))
        self.assertTrue(isinstance(self.server_stream, SSLIOStream))
        yield self.client_send_line(b"EHLO mail.example.com\r\n")
        yield self.server_send_line(b"250 mail.example.com welcome\r\n")

    @gen_test
    def test_handshake_fail(self):
        server_future = self.server_start_tls(_server_ssl_options())
        # Certificates are verified with the default configuration.
        client_future = self.client_start_tls(server_hostname="localhost")
        with ExpectLog(gen_log, "SSL Error"):
            with self.assertRaises(ssl.SSLError):
                yield client_future
        with self.assertRaises((ssl.SSLError, socket.error)):
            yield server_future

    @unittest.skipIf(not hasattr(ssl, 'create_default_context'),
                     'ssl.create_default_context not present')
    @gen_test
    def test_check_hostname(self):
        # Test that server_hostname parameter to start_tls is being used.
        # The check_hostname functionality is only available in python 2.7 and
        # up and in python 3.4 and up.
        server_future = self.server_start_tls(_server_ssl_options())
        client_future = self.client_start_tls(
            ssl.create_default_context(),
            server_hostname='127.0.0.1')
        with ExpectLog(gen_log, "SSL Error"):
            with self.assertRaises(ssl.SSLError):
                # The client fails to connect with an SSL error.
                yield client_future
        with self.assertRaises(Exception):
            # The server fails to connect, but the exact error is unspecified.
            yield server_future


class WaitForHandshakeTest(AsyncTestCase):
    @gen.coroutine
    def connect_to_server(self, server_cls):
        server = client = None
        try:
            sock, port = bind_unused_port()
            server = server_cls(ssl_options=_server_ssl_options())
            server.add_socket(sock)

            client = SSLIOStream(socket.socket(),
                                 ssl_options=dict(cert_reqs=ssl.CERT_NONE))
            yield client.connect(('127.0.0.1', port))
            self.assertIsNotNone(client.socket.cipher())
        finally:
            if server is not None:
                server.stop()
            if client is not None:
                client.close()

    @gen_test
    def test_wait_for_handshake_callback(self):
        test = self
        handshake_future = Future()

        class TestServer(TCPServer):
            def handle_stream(self, stream, address):
                # The handshake has not yet completed.
                test.assertIsNone(stream.socket.cipher())
                self.stream = stream
                stream.wait_for_handshake(self.handshake_done)

            def handshake_done(self):
                # Now the handshake is done and ssl information is available.
                test.assertIsNotNone(self.stream.socket.cipher())
                handshake_future.set_result(None)

        yield self.connect_to_server(TestServer)
        yield handshake_future

    @gen_test
    def test_wait_for_handshake_future(self):
        test = self
        handshake_future = Future()

        class TestServer(TCPServer):
            def handle_stream(self, stream, address):
                test.assertIsNone(stream.socket.cipher())
                test.io_loop.spawn_callback(self.handle_connection, stream)

            @gen.coroutine
            def handle_connection(self, stream):
                yield stream.wait_for_handshake()
                handshake_future.set_result(None)

        yield self.connect_to_server(TestServer)
        yield handshake_future

    @gen_test
    def test_wait_for_handshake_already_waiting_error(self):
        test = self
        handshake_future = Future()

        class TestServer(TCPServer):
            def handle_stream(self, stream, address):
                stream.wait_for_handshake(self.handshake_done)
                test.assertRaises(RuntimeError, stream.wait_for_handshake)

            def handshake_done(self):
                handshake_future.set_result(None)

        yield self.connect_to_server(TestServer)
        yield handshake_future

    @gen_test
    def test_wait_for_handshake_already_connected(self):
        handshake_future = Future()

        class TestServer(TCPServer):
            def handle_stream(self, stream, address):
                self.stream = stream
                stream.wait_for_handshake(self.handshake_done)

            def handshake_done(self):
                self.stream.wait_for_handshake(self.handshake2_done)

            def handshake2_done(self):
                handshake_future.set_result(None)

        yield self.connect_to_server(TestServer)
        yield handshake_future


@skipIfNonUnix
class TestPipeIOStream(AsyncTestCase):
    def test_pipe_iostream(self):
        r, w = os.pipe()

        rs = PipeIOStream(r, io_loop=self.io_loop)
        ws = PipeIOStream(w, io_loop=self.io_loop)

        ws.write(b"hel")
        ws.write(b"lo world")

        rs.read_until(b' ', callback=self.stop)
        data = self.wait()
        self.assertEqual(data, b"hello ")

        rs.read_bytes(3, self.stop)
        data = self.wait()
        self.assertEqual(data, b"wor")

        ws.close()

        rs.read_until_close(self.stop)
        data = self.wait()
        self.assertEqual(data, b"ld")

        rs.close()

    def test_pipe_iostream_big_write(self):
        r, w = os.pipe()

        rs = PipeIOStream(r, io_loop=self.io_loop)
        ws = PipeIOStream(w, io_loop=self.io_loop)

        NUM_BYTES = 1048576

        # Write 1MB of data, which should fill the buffer
        ws.write(b"1" * NUM_BYTES)

        rs.read_bytes(NUM_BYTES, self.stop)
        data = self.wait()
        self.assertEqual(data, b"1" * NUM_BYTES)

        ws.close()
        rs.close()



"""
A script to start the CherryPy WSGI server

This is run by ``salt-api`` and started in a multiprocess.
"""

import logging
import os

# pylint: disable=C0103


# Import CherryPy without traceback so we can provide an intelligent log
# message in the __virtual__ function
try:
    import cherrypy

    cpy_error = None
except ImportError as exc:
    cpy_error = exc

__virtualname__ = os.path.abspath(__file__).rsplit(os.sep)[-2] or "rest_cherrypy"

logger = logging.getLogger(__virtualname__)
cpy_min = "3.2.2"


def __virtual__():
    short_name = __name__.rsplit(".")[-1]
    mod_opts = __opts__.get(short_name, {})

    if mod_opts:
        # User has a rest_cherrypy section in config; assume the user wants to
        # run the module and increase logging severity to be helpful

        # Everything looks good; return the module name
        if not cpy_error and "port" in mod_opts:
            return __virtualname__

        # CherryPy wasn't imported; explain why
        if cpy_error:
            from salt.utils.versions import LooseVersion as V

            if "cherrypy" in globals() and V(cherrypy.__version__) < V(cpy_min):
                error_msg = "Required version of CherryPy is {} or greater.".format(
                    cpy_min
                )
            else:
                error_msg = cpy_error

            logger.error(
                "Not loading '%s'. Error loading CherryPy: %s", __name__, error_msg
            )

        # Missing port config
        if "port" not in mod_opts:
            logger.error("Not loading '%s'. 'port' not specified in config", __name__)

    return False


def verify_certs(*args):
    """
    Sanity checking for the specified SSL certificates
    """
    msg = (
        "Could not find a certificate: {0}\n"
        "If you want to quickly generate a self-signed certificate, "
        "use the tls.create_self_signed_cert function in Salt"
    )

    for arg in args:
        if not os.path.exists(arg):
            raise Exception(msg.format(arg))


def start():
    """
    Start the server loop
    """
    from . import app

    root, apiopts, conf = app.get_app(__opts__)

    if not apiopts.get("disable_ssl", False):
        if "ssl_crt" not in apiopts or "ssl_key" not in apiopts:
            logger.error(
                "Not starting '%s'. Options 'ssl_crt' and "
                "'ssl_key' are required if SSL is not disabled.",
                __name__,
            )

            return None

        verify_certs(apiopts["ssl_crt"], apiopts["ssl_key"])

        cherrypy.server.ssl_module = "builtin"
        cherrypy.server.ssl_certificate = apiopts["ssl_crt"]
        cherrypy.server.ssl_private_key = apiopts["ssl_key"]
        if "ssl_chain" in apiopts.keys():
            cherrypy.server.ssl_certificate_chain = apiopts["ssl_chain"]

    cherrypy.quickstart(root, apiopts.get("root_prefix", "/"), conf)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging
import os

import pytest  # pylint: disable=unused-import
import salt.exceptions
import salt.state
import salt.utils.files
import salt.utils.platform

log = logging.getLogger(__name__)


@pytest.fixture
def root_dir(tmp_path):
    return str(tmp_path / "root_dir")


@pytest.fixture
def base_state_tree_dir(root_dir):
    return os.path.join(root_dir, "base_state_tree")


@pytest.fixture
def other_state_tree_dir(root_dir):
    return os.path.join(root_dir, "other_state_tree")


@pytest.fixture
def cache_dir(root_dir):
    return os.path.join(root_dir, "cachedir")


@pytest.fixture
def highstate(
    temp_salt_minion,
    temp_salt_master,
    root_dir,
    base_state_tree_dir,
    other_state_tree_dir,
    cache_dir,
):
    for dpath in (
        root_dir,
        base_state_tree_dir,
        other_state_tree_dir,
        cache_dir,
    ):
        if not os.path.isdir(dpath):
            os.makedirs(dpath)

    test_sls = """
    test state:
      test.succeed_without_changes:
        - name: test
        """

    with pytest.helpers.temp_file("test.sls", test_sls, other_state_tree_dir):

        opts = temp_salt_minion.config.copy()
        opts["root_dir"] = root_dir
        opts["state_events"] = False
        opts["id"] = "match"
        opts["file_client"] = "local"
        opts["file_roots"] = dict(
            base=[base_state_tree_dir], other=[other_state_tree_dir]
        )
        opts["cachedir"] = cache_dir
        opts["test"] = False

        opts.update(
            {
                "transport": "zeromq",
                "auth_tries": 1,
                "auth_timeout": 5,
                "master_ip": "127.0.0.1",
                "master_port": temp_salt_master.config["ret_port"],
                "master_uri": "tcp://127.0.0.1:{}".format(
                    temp_salt_master.config["ret_port"]
                ),
            }
        )

        _highstate = salt.state.HighState(opts)
        _highstate.push_active()

        yield _highstate


def test_lazy_avail_states_base(highstate, base_state_tree_dir, tmp_path):
    top_sls = """
    base:
      '*':
        - core
        """

    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        str(tmp_path)
    )

    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_state_tree_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_state_tree_dir):
        # list_states not called yet
        assert not highstate.avail._filled
        assert highstate.avail._avail == {"base": None}
        # After getting 'base' env available states
        highstate.avail["base"]  # pylint: disable=pointless-statement
        assert not highstate.avail._filled
        assert highstate.avail._avail == {"base": ["core", "top"]}


def test_lazy_avail_states_other(highstate, base_state_tree_dir, tmp_path):
    top_sls = """
    base:
      '*':
        - core
        """

    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        str(tmp_path)
    )

    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_state_tree_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_state_tree_dir):
        # list_states not called yet
        assert not highstate.avail._filled
        assert highstate.avail._avail == {"base": None}
        # After getting 'other' env available states
        highstate.avail["other"]  # pylint: disable=pointless-statement
        assert highstate.avail._filled
        assert highstate.avail._avail == {"base": None, "other": ["test"]}


def test_lazy_avail_states_multi(highstate, base_state_tree_dir, tmp_path):
    top_sls = """
    base:
      '*':
        - core
        """

    core_state = """
    {}/testfile:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        str(tmp_path)
    )

    with pytest.helpers.temp_file(
        "top.sls", top_sls, base_state_tree_dir
    ), pytest.helpers.temp_file("core.sls", core_state, base_state_tree_dir):
        # list_states not called yet
        assert not highstate.avail._filled
        assert highstate.avail._avail == {"base": None}
        # After getting 'base' env available states
        highstate.avail["base"]  # pylint: disable=pointless-statement
        assert not highstate.avail._filled
        assert highstate.avail._avail == {"base": ["core", "top"]}
        # After getting 'other' env available states
        highstate.avail["other"]  # pylint: disable=pointless-statement
        assert highstate.avail._filled
        assert highstate.avail._avail == {"base": ["core", "top"], "other": ["test"]}



"""
    :codeauthor: `Anthony Shaw <anthonyshaw@apache.org>`

    tests.unit.cloud.clouds.dimensiondata_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from salt.cloud.clouds import dimensiondata
from salt.exceptions import SaltCloudSystemExit
from salt.utils.versions import LooseVersion
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock
from tests.support.mock import __version__ as mock_version
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf
from tests.unit.cloud.clouds import _preferred_ip

try:
    import libcloud.security

    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False


VM_NAME = "winterfell"

# Use certifi if installed
try:
    if HAS_LIBCLOUD:
        # This work-around for Issue #32743 is no longer needed for libcloud >=
        # 1.4.0. However, older versions of libcloud must still be supported
        # with this work-around. This work-around can be removed when the
        # required minimum version of libcloud is 2.0.0 (See PR #40837 - which
        # is implemented in Salt 2018.3.0).
        if LooseVersion(libcloud.__version__) < LooseVersion("1.4.0"):
            import certifi

            libcloud.security.CA_CERTS_PATH.append(certifi.where())
except (ImportError, NameError):
    pass


class ExtendedTestCase(TestCase):
    """
    Extended TestCase class containing additional helper methods.
    """

    def assertRaisesWithMessage(self, exc_type, exc_msg, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
            self.assertFail()
        except Exception as exc:  # pylint: disable=broad-except
            self.assertEqual(type(exc), exc_type)
            self.assertEqual(exc.message, exc_msg)


class DimensionDataTestCase(ExtendedTestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for salt.cloud.clouds.dimensiondata module.
    """

    def setup_loader_modules(self):
        return {
            dimensiondata: {
                "__active_provider_name__": "",
                "__opts__": {
                    "providers": {
                        "my-dimensiondata-cloud": {
                            "dimensiondata": {
                                "driver": "dimensiondata",
                                "region": "dd-au",
                                "user_id": "jon_snow",
                                "key": "IKnowNothing",
                            }
                        }
                    }
                },
            }
        }

    def test_avail_images_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call avail_images
        with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, dimensiondata.avail_images, call="action"
        )

    def test_avail_locations_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call avail_locations
        with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, dimensiondata.avail_locations, call="action"
        )

    def test_avail_sizes_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call avail_sizes
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, dimensiondata.avail_sizes, call="action")

    def test_list_nodes_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_nodes
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, dimensiondata.list_nodes, call="action")

    def test_destroy_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call destroy
        with --function or -f.
        """
        self.assertRaises(
            SaltCloudSystemExit, dimensiondata.destroy, name=VM_NAME, call="function"
        )

    @skipIf(
        HAS_LIBCLOUD is False, "Install 'libcloud' to be able to run this unit test."
    )
    def test_avail_sizes(self):
        """
        Tests that avail_sizes returns an empty dictionary.
        """
        sizes = dimensiondata.avail_sizes(call="foo")
        self.assertEqual(len(sizes), 1)
        self.assertEqual(sizes["default"]["name"], "default")

    def test_import(self):
        """
        Test that the module picks up installed deps
        """
        with patch("salt.config.check_driver_dependencies", return_value=True) as p:
            get_deps = dimensiondata.get_dependencies()
            self.assertEqual(get_deps, True)
            if LooseVersion(mock_version) >= LooseVersion("2.0.0"):
                self.assertTrue(p.call_count >= 1)

    def test_provider_matches(self):
        """
        Test that the first configured instance of a dimensiondata driver is matched
        """
        p = dimensiondata.get_configured_provider()
        self.assertNotEqual(p, None)

    def test_query_node_data_filter_preferred_ip_addresses(self):
        """
        Test if query node data is filtering out unpreferred IP addresses.
        """
        zero_ip = "0.0.0.0"
        private_ips = [zero_ip, "1.1.1.1", "2.2.2.2"]
        vm = {"name": None}
        data = MagicMock()
        data.public_ips = []
        # pylint: disable=blacklisted-unmocked-patching
        dimensiondata.NodeState = MagicMock()
        # pylint: enable=blacklisted-unmocked-patching
        dimensiondata.NodeState.RUNNING = True

        with patch(
            "salt.cloud.clouds.dimensiondata.show_instance",
            MagicMock(
                return_value={
                    "state": True,
                    "name": "foo",
                    "public_ips": [],
                    "private_ips": private_ips,
                }
            ),
        ):
            with patch(
                "salt.cloud.clouds.dimensiondata.preferred_ip",
                _preferred_ip(private_ips, [zero_ip]),
            ):
                with patch(
                    "salt.cloud.clouds.dimensiondata.ssh_interface",
                    MagicMock(return_value="private_ips"),
                ):
                    self.assertEqual(
                        dimensiondata._query_node_data(vm, data).public_ips, [zero_ip]
                    )



"""
tests.unit.proxy.test_cimc
~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the cimc proxy module
"""

import logging

import salt.exceptions
import salt.proxy.cimc as cimc
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


LOGIN_RESPONSE = """\
<aaaLogin
    response="yes"
    outCookie="real-cookie"
       outRefreshPeriod="600"
       outPriv="admin">
</aaaLogin>"""
LOGOUT_RESPONSE = """\
<aaaLogout
    cookie="real-cookie"
    response="yes"
    outStatus="success">
</aaaLogout>
"""
CONFIG_RESOLVE_CLASS_RESPONSE = """\
<configResolveClass
    cookie="real-cookie"
    response="yes"
    classId="computeRackUnit">
    <outConfig>
        <computeRackUnit
            dn="sys/rack-unit-1"
            adminPower="policy"
            availableMemory="16384"
            model="R210-2121605W"
            memorySpeed="1067"
            name="UCS C210 M2"
            numOfAdaptors="2"
            numOfCores="8"
            numOfCoresEnabled="8"
            numOfCpus="2"
            numOfEthHostIfs="5"
            numOfFcHostIfs="2"
            numOfThreads="16"
            operPower="on"
            originalUuid="00C9DE3C-370D-DF11-1186-6DD1393A608B"
            presence="equipped"
            serverID="1"
            serial="QCI140205Z2"
            totalMemory="16384"
            usrLbl="C210 Row-B Rack-10"
            uuid="00C9DE3C-370D-DF11-1186-6DD1393A608B"
            vendor="Cisco Systems Inc" >
        </computeRackUnit>
    </outConfig>
</configResolveClass>
"""
CONFIG_CON_MO_RESPONSE = """\
<configConfMo
    dn="sys/rack-unit-1/locator-led"
    cookie="real-cookie"
    response="yes">
    <outConfig>
        <equipmentLocatorLed
            dn="sys/rack-unit-1/locator-led"
            adminState="inactive"
            color="unknown"
            id="1"
            name=""
            operState="off">
        </equipmentLocatorLed>
    </outConfig>
</configConfMo>
"""


def http_query_response(*args, data=None, **kwargs):
    log.debug(
        "http_query_response side_effect; ARGS: %s // KWARGS: %s // DATA: %s",
        args,
        kwargs,
        data,
    )
    if data.startswith("<aaaLogin"):
        response = LOGIN_RESPONSE
    elif data.startswith("<aaaLogout"):
        response = LOGOUT_RESPONSE
    elif data.startswith("<configResolveClass"):
        response = CONFIG_RESOLVE_CLASS_RESPONSE
    elif data.startswith("<configConfMo"):
        response = CONFIG_CON_MO_RESPONSE
    else:
        response = ""
    return {"text": response, "status": 200}


class CIMCProxyTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {cimc: {"DETAILS": {}, "__pillar__": {}}}

    def setUp(self):
        self.opts = {"proxy": {"username": "xxxx", "password": "xxx", "host": "cimc"}}
        self.addCleanup(delattr, self, "opts")

    def test_init(self):
        # No host, returns False
        opts = {"proxy": {"username": "xxxx", "password": "xxx"}}
        ret = cimc.init(opts)
        self.assertFalse(ret)

        # No username , returns False
        opts = {"proxy": {"password": "xxx", "host": "cimc"}}
        ret = cimc.init(opts)
        self.assertFalse(ret)

        # No password, returns False
        opts = {"proxy": {"username": "xxxx", "host": "cimc"}}
        ret = cimc.init(opts)
        self.assertFalse(ret)

        with patch.object(cimc, "logon", return_value="9zVG5U8DFZNsTR") as mock_logon:
            with patch.object(
                cimc, "get_config_resolver_class", return_value="True"
            ) as mock_logon:
                ret = cimc.init(self.opts)
                self.assertEqual(cimc.DETAILS["url"], "https://cimc/nuova")
                self.assertEqual(cimc.DETAILS["username"], "xxxx")
                self.assertEqual(cimc.DETAILS["password"], "xxx")
                self.assertTrue(cimc.DETAILS["initialized"])

    def test__validate_response_code(self):
        with self.assertRaisesRegex(
            salt.exceptions.CommandExecutionError,
            "Did not receive a valid response from host.",
        ):
            cimc._validate_response_code("404")

        with patch.object(cimc, "logout", return_value=True) as mock_logout:
            with self.assertRaisesRegex(
                salt.exceptions.CommandExecutionError,
                "Did not receive a valid response from host.",
            ):
                cimc._validate_response_code("404", "9zVG5U8DFZNsTR")
                mock_logout.assert_called_once_with("9zVG5U8DFZNsTR")


class ValidateSSLTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {cimc: {}}

    def setUp(self):
        cimc.DETAILS.clear()

    def test_init(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.init(opts)

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_logon(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.logon()

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_logout(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.logout()

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_grains(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.grains()

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_grains_refresh(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.grains_refresh()

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_ping(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.ping()

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)

    def test_set_config_modify(self):
        verify_ssl_values = (None, True, False)
        for verify_ssl_value in verify_ssl_values:
            cimc.DETAILS.clear()
            opts = {
                "proxy": {
                    "host": "TheHost",
                    "username": "TheUsername",
                    "password": "ThePassword",
                    "verify_ssl": verify_ssl_value,
                }
            }
            http_query_mock = MagicMock(side_effect=http_query_response)
            if verify_ssl_value is None:
                expected_verify_ssl_value = True
            else:
                expected_verify_ssl_value = verify_ssl_value

            # Let's init the proxy and ignore it's actions, this test is not about them
            with patch(
                "salt.proxy.cimc.get_config_resolver_class",
                MagicMock(return_value=True),
            ):
                cimc.init(opts)

            log.debug(
                "verify_ssl: %s // expected verify_ssl: %s",
                verify_ssl_value,
                expected_verify_ssl_value,
            )

            with patch.dict(cimc.__utils__, {"http.query": http_query_mock}):
                cimc.set_config_modify(
                    dn="sys/rack-unit-1/locator-led",
                    inconfig=(
                        "<inConfig><equipmentLocatorLed adminState='on'"
                        " dn='sys/rack-unit-1/locator-led'></equipmentLocatorLed></inConfig>"
                    ),
                )

            for idx, call in enumerate(http_query_mock.mock_calls, 1):
                condition = call.kwargs["verify_ssl"] is expected_verify_ssl_value
                condition_error = "{} != {}; Call(number={}): {}".format(
                    idx, call, call.kwargs["verify_ssl"], expected_verify_ssl_value
                )
                self.assertTrue(condition, msg=condition_error)



import pytest
import salt.utils.stringutils
import salt.utils.win_reg as win_reg
from salt.exceptions import CommandExecutionError
from tests.support.helpers import random_string
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    import win32api

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

UNICODE_KEY = "Unicode Key \N{TRADE MARK SIGN}"
UNICODE_VALUE = (
    "Unicode Value \N{COPYRIGHT SIGN},\N{TRADE MARK SIGN},\N{REGISTERED SIGN}"
)
FAKE_KEY = "SOFTWARE\\{}".format(random_string("SaltTesting-", lowercase=False))


@skipIf(not HAS_WIN32, "Tests require win32 libraries")
class WinFunctionsTestCase(TestCase):
    """
    Test cases for salt.utils.win_reg
    """

    def test_broadcast_change_success(self):
        """
        Tests the broadcast_change function
        """
        with patch("win32gui.SendMessageTimeout", return_value=("", 0)):
            self.assertTrue(win_reg.broadcast_change())

    def test_broadcast_change_fail(self):
        """
        Tests the broadcast_change function failure
        """
        with patch("win32gui.SendMessageTimeout", return_value=("", 1)):
            self.assertFalse(win_reg.broadcast_change())

    def test_key_exists_existing(self):
        """
        Tests the key_exists function using a well known registry key
        """
        self.assertTrue(win_reg.key_exists(hive="HKLM", key="SOFTWARE\\Microsoft"))

    def test_key_exists_non_existing(self):
        """
        Tests the key_exists function using a non existing registry key
        """
        self.assertFalse(win_reg.key_exists(hive="HKLM", key=FAKE_KEY))

    def test_key_exists_invalid_hive(self):
        """
        Tests the key_exists function using an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.key_exists,
            hive="BADHIVE",
            key="SOFTWARE\\Microsoft",
        )

    def test_key_exists_unknown_key_error(self):
        """
        Tests the key_exists function with an unknown key error
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.key_exists,
                hive="HKLM",
                key="SOFTWARE\\Microsoft",
            )

    def test_value_exists_existing(self):
        """
        Tests the value_exists function using a well known registry key
        """
        self.assertTrue(
            win_reg.value_exists(
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="CommonFilesDir",
            )
        )

    def test_value_exists_non_existing(self):
        """
        Tests the value_exists function using a non existing registry key
        """
        self.assertFalse(
            win_reg.value_exists(
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="NonExistingValueName",
            )
        )

    def test_value_exists_invalid_hive(self):
        """
        Tests the value_exists function using an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.value_exists,
            hive="BADHIVE",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="CommonFilesDir",
        )

    def test_value_exists_key_not_exist(self):
        """
        Tests the value_exists function when the key does not exist
        """
        mock_error = MagicMock(
            side_effect=win32api.error(2, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertFalse(
                win_reg.value_exists(
                    hive="HKLM",
                    key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                    vname="CommonFilesDir",
                )
            )

    def test_value_exists_unknown_key_error(self):
        """
        Tests the value_exists function with an unknown error when opening the
        key
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.value_exists,
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="CommonFilesDir",
            )

    def test_value_exists_empty_default_value(self):
        """
        Tests the value_exists function when querying the default value
        """
        mock_error = MagicMock(
            side_effect=win32api.error(2, "RegQueryValueEx", "Empty Value")
        )
        with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
            self.assertTrue(
                win_reg.value_exists(
                    hive="HKLM",
                    key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                    vname=None,
                )
            )

    def test_value_exists_no_vname(self):
        """
        Tests the value_exists function when the vname does not exist
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegQueryValueEx", "Empty Value")
        )
        with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
            self.assertFalse(
                win_reg.value_exists(
                    hive="HKLM",
                    key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                    vname="NonExistingValuePair",
                )
            )

    def test_list_keys_existing(self):
        """
        Test the list_keys function using a well known registry key
        """
        self.assertIn("Microsoft", win_reg.list_keys(hive="HKLM", key="SOFTWARE"))

    def test_list_keys_non_existing(self):
        """
        Test the list_keys function using a non existing registry key
        """
        expected = (False, "Cannot find key: HKLM\\{}".format(FAKE_KEY))
        self.assertEqual(win_reg.list_keys(hive="HKLM", key=FAKE_KEY), expected)

    def test_list_keys_invalid_hive(self):
        """
        Test the list_keys function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.list_keys,
            hive="BADHIVE",
            key="SOFTWARE\\Microsoft",
        )

    def test_list_keys_unknown_key_error(self):
        """
        Tests the list_keys function with an unknown key error
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.list_keys,
                hive="HKLM",
                key="SOFTWARE\\Microsoft",
            )

    def test_list_values_existing(self):
        """
        Test the list_values function using a well known registry key
        """
        values = win_reg.list_values(
            hive="HKLM", key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
        )
        keys = []
        for value in values:
            keys.append(value["vname"])
        self.assertIn("ProgramFilesDir", keys)

    def test_list_values_non_existing(self):
        """
        Test the list_values function using a non existing registry key
        """
        expected = (False, "Cannot find key: HKLM\\{}".format(FAKE_KEY))
        self.assertEqual(win_reg.list_values(hive="HKLM", key=FAKE_KEY), expected)

    def test_list_values_invalid_hive(self):
        """
        Test the list_values function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.list_values,
            hive="BADHIVE",
            key="SOFTWARE\\Microsoft",
        )

    def test_list_values_unknown_key_error(self):
        """
        Tests the list_values function with an unknown key error
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.list_values,
                hive="HKLM",
                key="SOFTWARE\\Microsoft",
            )

    def test_read_value_existing(self):
        """
        Test the read_value function using a well known registry value
        """
        ret = win_reg.read_value(
            hive="HKLM",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
            vname="ProgramFilesPath",
        )
        self.assertEqual(ret["vdata"], "%ProgramFiles%")

    def test_read_value_default(self):
        """
        Test the read_value function reading the default value using a well
        known registry key
        """
        ret = win_reg.read_value(
            hive="HKLM", key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion"
        )
        self.assertEqual(ret["vdata"], "(value not set)")

    def test_read_value_non_existing(self):
        """
        Test the read_value function using a non existing value pair
        """
        expected = {
            "comment": (
                "Cannot find fake_name in HKLM\\SOFTWARE\\Microsoft\\"
                "Windows\\CurrentVersion"
            ),
            "vdata": None,
            "vname": "fake_name",
            "success": False,
            "hive": "HKLM",
            "key": "SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
        }
        self.assertDictEqual(
            win_reg.read_value(
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="fake_name",
            ),
            expected,
        )

    def test_read_value_non_existing_key(self):
        """
        Test the read_value function using a non existing registry key
        """
        expected = {
            "comment": "Cannot find key: HKLM\\{}".format(FAKE_KEY),
            "vdata": None,
            "vname": "fake_name",
            "success": False,
            "hive": "HKLM",
            "key": FAKE_KEY,
        }
        self.assertDictEqual(
            win_reg.read_value(hive="HKLM", key=FAKE_KEY, vname="fake_name"), expected
        )

    def test_read_value_invalid_hive(self):
        """
        Test the read_value function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.read_value,
            hive="BADHIVE",
            key="SOFTWARE\\Microsoft",
            vname="ProgramFilesPath",
        )

    def test_read_value_unknown_key_error(self):
        """
        Tests the read_value function with an unknown key error
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.read_value,
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="ProgramFilesPath",
            )

    def test_read_value_unknown_value_error(self):
        """
        Tests the read_value function with an unknown value error
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegQueryValueEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegQueryValueEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.read_value,
                hive="HKLM",
                key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
                vname="ProgramFilesPath",
            )

    @pytest.mark.destructive_test
    def test_read_value_multi_sz_empty_list(self):
        """
        An empty REG_MULTI_SZ value should return an empty list, not None
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key=FAKE_KEY,
                    vname="empty_list",
                    vdata=[],
                    vtype="REG_MULTI_SZ",
                )
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": [],
                "vname": "empty_list",
                "vtype": "REG_MULTI_SZ",
            }
            self.assertEqual(
                win_reg.read_value(
                    hive="HKLM",
                    key=FAKE_KEY,
                    vname="empty_list",
                ),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value(self):
        """
        Test the set_value function
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": "fake_data",
                "vname": "fake_name",
                "vtype": "REG_SZ",
            }
            self.assertEqual(
                win_reg.read_value(hive="HKLM", key=FAKE_KEY, vname="fake_name"),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value_default(self):
        """
        Test the set_value function on the default value
        """
        try:
            self.assertTrue(
                win_reg.set_value(hive="HKLM", key=FAKE_KEY, vdata="fake_default_data")
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": "fake_default_data",
                "vname": "(Default)",
                "vtype": "REG_SZ",
            }
            self.assertEqual(win_reg.read_value(hive="HKLM", key=FAKE_KEY), expected)
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value_unicode_key(self):
        """
        Test the set_value function on a unicode key
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key="\\".join([FAKE_KEY, UNICODE_KEY]),
                    vname="fake_name",
                    vdata="fake_value",
                )
            )
            expected = {
                "hive": "HKLM",
                "key": "\\".join([FAKE_KEY, UNICODE_KEY]),
                "success": True,
                "vdata": "fake_value",
                "vname": "fake_name",
                "vtype": "REG_SZ",
            }
            self.assertEqual(
                win_reg.read_value(
                    hive="HKLM",
                    key="\\".join([FAKE_KEY, UNICODE_KEY]),
                    vname="fake_name",
                ),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value_unicode_value(self):
        """
        Test the set_value function on a unicode value
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_unicode", vdata=UNICODE_VALUE
                )
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": UNICODE_VALUE,
                "vname": "fake_unicode",
                "vtype": "REG_SZ",
            }
            self.assertEqual(
                win_reg.read_value(hive="HKLM", key=FAKE_KEY, vname="fake_unicode"),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value_reg_dword(self):
        """
        Test the set_value function on a REG_DWORD value
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key=FAKE_KEY,
                    vname="dword_value",
                    vdata=123,
                    vtype="REG_DWORD",
                )
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": 123,
                "vname": "dword_value",
                "vtype": "REG_DWORD",
            }
            self.assertEqual(
                win_reg.read_value(hive="HKLM", key=FAKE_KEY, vname="dword_value"),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_set_value_reg_qword(self):
        """
        Test the set_value function on a REG_QWORD value
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key=FAKE_KEY,
                    vname="qword_value",
                    vdata=123,
                    vtype="REG_QWORD",
                )
            )
            expected = {
                "hive": "HKLM",
                "key": FAKE_KEY,
                "success": True,
                "vdata": 123,
                "vname": "qword_value",
                "vtype": "REG_QWORD",
            }
            self.assertEqual(
                win_reg.read_value(hive="HKLM", key=FAKE_KEY, vname="qword_value"),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    def test_set_value_invalid_hive(self):
        """
        Test the set_value function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.set_value,
            hive="BADHIVE",
            key=FAKE_KEY,
            vname="fake_name",
            vdata="fake_data",
        )

    def test_set_value_open_create_failure(self):
        """
        Test the set_value function when there is a problem opening/creating
        the key
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegCreateKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegCreateKeyEx", mock_error):
            self.assertFalse(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )

    def test_set_value_type_error(self):
        """
        Test the set_value function when the wrong type of data is passed
        """
        mock_error = MagicMock(side_effect=TypeError("Mocked TypeError"))
        with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
            self.assertFalse(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )

    def test_set_value_system_error(self):
        """
        Test the set_value function when a SystemError occurs while setting the
        value
        """
        mock_error = MagicMock(side_effect=SystemError("Mocked SystemError"))
        with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
            self.assertFalse(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )

    def test_set_value_value_error(self):
        """
        Test the set_value function when a ValueError occurs while setting the
        value
        """
        mock_error = MagicMock(side_effect=ValueError("Mocked ValueError"))
        with patch("salt.utils.win_reg.win32api.RegSetValueEx", mock_error):
            self.assertFalse(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )

    def test_cast_vdata_reg_binary(self):
        """
        Test the cast_vdata function with REG_BINARY
        Should always return binary data
        """
        vdata = salt.utils.stringutils.to_bytes("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
        self.assertTrue(isinstance(result, bytes))

        vdata = salt.utils.stringutils.to_str("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
        self.assertTrue(isinstance(result, bytes))

        vdata = salt.utils.stringutils.to_unicode("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_BINARY")
        self.assertTrue(isinstance(result, bytes))

    def test_cast_vdata_reg_dword(self):
        """
        Test the cast_vdata function with REG_DWORD
        Should always return integer
        """
        vdata = 1
        expected = 1
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
        self.assertEqual(result, expected)

        vdata = "1"
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
        self.assertEqual(result, expected)

        vdata = "0000001"
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_DWORD")
        self.assertEqual(result, expected)

    def test_cast_vdata_reg_expand_sz(self):
        """
        Test the cast_vdata function with REG_EXPAND_SZ
        Should always return unicode
        """
        vdata = salt.utils.stringutils.to_str("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_EXPAND_SZ")
        self.assertTrue(isinstance(result, str))

        vdata = salt.utils.stringutils.to_bytes("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_EXPAND_SZ")
        self.assertTrue(isinstance(result, str))

    def test_cast_vdata_reg_multi_sz(self):
        """
        Test the cast_vdata function with REG_MULTI_SZ
        Should always return a list of unicode strings
        """
        vdata = [
            salt.utils.stringutils.to_str("test string"),
            salt.utils.stringutils.to_bytes("test bytes"),
        ]
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_MULTI_SZ")
        self.assertTrue(isinstance(result, list))
        for item in result:
            self.assertTrue(isinstance(item, str))

    def test_cast_vdata_reg_qword(self):
        """
        Test the cast_vdata function with REG_QWORD
        Should always return a long integer
        `int` is `long` is default on Py3
        """
        vdata = 1
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_QWORD")
        self.assertTrue(isinstance(result, int))

        vdata = "1"
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_QWORD")
        self.assertTrue(isinstance(result, int))

    def test_cast_vdata_reg_sz(self):
        """
        Test the cast_vdata function with REG_SZ
        Should always return unicode
        """
        vdata = salt.utils.stringutils.to_str("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
        self.assertTrue(isinstance(result, str))

        vdata = salt.utils.stringutils.to_bytes("test data")
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
        self.assertTrue(isinstance(result, str))

        vdata = None
        result = win_reg.cast_vdata(vdata=vdata, vtype="REG_SZ")
        self.assertTrue(isinstance(result, str))
        self.assertEqual(result, "")

    @pytest.mark.destructive_test
    def test_delete_value(self):
        """
        Test the delete_value function
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_data"
                )
            )
            self.assertTrue(
                win_reg.delete_value(hive="HKLM", key=FAKE_KEY, vname="fake_name")
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    def test_delete_value_non_existing(self):
        """
        Test the delete_value function on non existing value
        """
        mock_error = MagicMock(
            side_effect=win32api.error(2, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertIsNone(
                win_reg.delete_value(hive="HKLM", key=FAKE_KEY, vname="fake_name")
            )

    def test_delete_value_invalid_hive(self):
        """
        Test the delete_value function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.delete_value,
            hive="BADHIVE",
            key=FAKE_KEY,
            vname="fake_name",
        )

    def test_delete_value_unknown_error(self):
        """
        Test the delete_value function when there is a problem opening the key
        """
        mock_error = MagicMock(
            side_effect=win32api.error(123, "RegOpenKeyEx", "Unknown error")
        )
        with patch("salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error):
            self.assertRaises(
                win32api.error,
                win_reg.delete_value,
                hive="HKLM",
                key=FAKE_KEY,
                vname="fake_name",
            )

    @pytest.mark.destructive_test
    def test_delete_value_unicode(self):
        """
        Test the delete_value function on a unicode value
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_unicode", vdata=UNICODE_VALUE
                )
            )
            self.assertTrue(
                win_reg.delete_value(hive="HKLM", key=FAKE_KEY, vname="fake_unicode")
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_delete_value_unicode_vname(self):
        """
        Test the delete_value function on a unicode vname
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname=UNICODE_KEY, vdata="junk data"
                )
            )
            self.assertTrue(
                win_reg.delete_value(hive="HKLM", key=FAKE_KEY, vname=UNICODE_KEY)
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_delete_value_unicode_key(self):
        """
        Test the delete_value function on a unicode key
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key="\\".join([FAKE_KEY, UNICODE_KEY]),
                    vname="fake_name",
                    vdata="junk data",
                )
            )
            self.assertTrue(
                win_reg.delete_value(
                    hive="HKLM",
                    key="\\".join([FAKE_KEY, UNICODE_KEY]),
                    vname="fake_name",
                )
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    def test_delete_key_recursive_invalid_hive(self):
        """
        Test the delete_key_recursive function when passing an invalid hive
        """
        self.assertRaises(
            CommandExecutionError,
            win_reg.delete_key_recursive,
            hive="BADHIVE",
            key=FAKE_KEY,
        )

    def test_delete_key_recursive_key_not_found(self):
        """
        Test the delete_key_recursive function when the passed key to delete is
        not found.
        """
        self.assertFalse(win_reg.key_exists(hive="HKLM", key=FAKE_KEY))
        self.assertFalse(win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY))

    def test_delete_key_recursive_too_close(self):
        """
        Test the delete_key_recursive function when the passed key to delete is
        too close to root, such as
        """
        mock_true = MagicMock(return_value=True)
        with patch("salt.utils.win_reg.key_exists", mock_true):
            self.assertFalse(win_reg.delete_key_recursive(hive="HKLM", key="FAKE_KEY"))

    @pytest.mark.destructive_test
    def test_delete_key_recursive(self):
        """
        Test the delete_key_recursive function
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_value"
                )
            )
            expected = {"Deleted": ["\\".join(["HKLM", FAKE_KEY])], "Failed": []}
            self.assertDictEqual(
                win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY), expected
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_delete_key_recursive_failed_to_open_key(self):
        """
        Test the delete_key_recursive function on failure to open the key
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_value"
                )
            )
            expected = {
                "Deleted": [],
                "Failed": ["\\".join(["HKLM", FAKE_KEY]) + " Failed to connect to key"],
            }
            mock_true = MagicMock(return_value=True)
            mock_error = MagicMock(
                side_effect=[
                    1,
                    win32api.error(3, "RegOpenKeyEx", "Failed to connect to key"),
                ]
            )
            with patch("salt.utils.win_reg.key_exists", mock_true), patch(
                "salt.utils.win_reg.win32api.RegOpenKeyEx", mock_error
            ):
                self.assertDictEqual(
                    win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY), expected
                )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_delete_key_recursive_failed_to_delete(self):
        """
        Test the delete_key_recursive function on failure to delete a key
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM", key=FAKE_KEY, vname="fake_name", vdata="fake_value"
                )
            )
            expected = {
                "Deleted": [],
                "Failed": ["\\".join(["HKLM", FAKE_KEY]) + " Unknown error"],
            }
            # pylint: disable=undefined-variable
            mock_error = MagicMock(side_effect=WindowsError("Unknown error"))
            # pylint: enable=undefined-variable
            with patch("salt.utils.win_reg.win32api.RegDeleteKey", mock_error):
                self.assertDictEqual(
                    win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY), expected
                )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    @pytest.mark.destructive_test
    def test_delete_key_recursive_unicode(self):
        """
        Test the delete_key_recursive function on value within a unicode key
        """
        try:
            self.assertTrue(
                win_reg.set_value(
                    hive="HKLM",
                    key="\\".join([FAKE_KEY, UNICODE_KEY]),
                    vname="fake_name",
                    vdata="fake_value",
                )
            )
            expected = {
                "Deleted": ["\\".join(["HKLM", FAKE_KEY, UNICODE_KEY])],
                "Failed": [],
            }
            self.assertDictEqual(
                win_reg.delete_key_recursive(
                    hive="HKLM", key="\\".join([FAKE_KEY, UNICODE_KEY])
                ),
                expected,
            )
        finally:
            win_reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)

    def test__to_unicode_int(self):
        """
        Test the ``_to_unicode`` function when it receives an integer value.
        Should return a unicode value, which is unicode in PY2 and str in PY3.
        """
        self.assertTrue(isinstance(win_reg._to_unicode(1), str))



"""
:maintainer:    SaltStack
:maturity:      new
:platform:      all

Runner functions supporting the Vault modules. Configuration instructions are
documented in the execution module docs.
"""

import base64
import json
import logging
import string
import time

import requests
import salt.crypt
import salt.exceptions

log = logging.getLogger(__name__)


def generate_token(
    minion_id, signature, impersonated_by_master=False, ttl=None, uses=None
):
    """
    Generate a Vault token for minion minion_id

    minion_id
        The id of the minion that requests a token

    signature
        Cryptographic signature which validates that the request is indeed sent
        by the minion (or the master, see impersonated_by_master).

    impersonated_by_master
        If the master needs to create a token on behalf of the minion, this is
        True. This happens when the master generates minion pillars.

    ttl
        Ticket time to live in seconds, 1m minutes, or 2h hrs

    uses
        Number of times a token can be used
    """
    log.debug(
        "Token generation request for %s (impersonated by master: %s)",
        minion_id,
        impersonated_by_master,
    )
    _validate_signature(minion_id, signature, impersonated_by_master)
    try:
        config = __opts__.get("vault", {})
        verify = config.get("verify", None)
        # Vault Enterprise requires a namespace
        namespace = config.get("namespace")
        # Allow disabling of minion provided values via the master
        allow_minion_override = config["auth"].get("allow_minion_override", False)
        # This preserves the previous behavior of default TTL and 1 use
        if not allow_minion_override or uses is None:
            uses = config["auth"].get("uses", 1)
        if not allow_minion_override or ttl is None:
            ttl = config["auth"].get("ttl", None)
        storage_type = config["auth"].get("token_backend", "session")

        if config["auth"]["method"] == "approle":
            if _selftoken_expired():
                log.debug("Vault token expired. Recreating one")
                # Requesting a short ttl token
                url = "{}/v1/auth/approle/login".format(config["url"])
                payload = {"role_id": config["auth"]["role_id"]}
                if "secret_id" in config["auth"]:
                    payload["secret_id"] = config["auth"]["secret_id"]
                # Vault Enterprise call requires headers
                headers = None
                if namespace is not None:
                    headers = {"X-Vault-Namespace": namespace}
                response = requests.post(
                    url, headers=headers, json=payload, verify=verify
                )
                if response.status_code != 200:
                    return {"error": response.reason}
                config["auth"]["token"] = response.json()["auth"]["client_token"]

        url = _get_token_create_url(config)
        headers = {"X-Vault-Token": config["auth"]["token"]}
        if namespace is not None:
            headers["X-Vault-Namespace"] = namespace
        audit_data = {
            "saltstack-jid": globals().get("__jid__", "<no jid set>"),
            "saltstack-minion": minion_id,
            "saltstack-user": globals().get("__user__", "<no user set>"),
        }
        payload = {
            "policies": _get_policies(minion_id, config),
            "num_uses": uses,
            "meta": audit_data,
        }

        if ttl is not None:
            payload["explicit_max_ttl"] = str(ttl)

        if payload["policies"] == []:
            return {"error": "No policies matched minion"}

        log.trace("Sending token creation request to Vault")
        response = requests.post(url, headers=headers, json=payload, verify=verify)

        if response.status_code != 200:
            return {"error": response.reason}

        auth_data = response.json()["auth"]
        ret = {
            "token": auth_data["client_token"],
            "lease_duration": auth_data["lease_duration"],
            "renewable": auth_data["renewable"],
            "issued": int(round(time.time())),
            "url": config["url"],
            "verify": verify,
            "token_backend": storage_type,
            "namespace": namespace,
        }
        if uses >= 0:
            ret["uses"] = uses

        return ret
    except Exception as e:  # pylint: disable=broad-except
        return {"error": str(e)}


def unseal():
    """
    Unseal Vault server

    This function uses the 'keys' from the 'vault' configuration to unseal vault server

    vault:
      keys:
        - n63/TbrQuL3xaIW7ZZpuXj/tIfnK1/MbVxO4vT3wYD2A
        - S9OwCvMRhErEA4NVVELYBs6w/Me6+urgUr24xGK44Uy3
        - F1j4b7JKq850NS6Kboiy5laJ0xY8dWJvB3fcwA+SraYl
        - 1cYtvjKJNDVam9c7HNqJUfINk4PYyAXIpjkpN/sIuzPv
        - 3pPK5X6vGtwLhNOFv1U2elahECz3HpRUfNXJFYLw6lid

    .. note: This function will send unsealed keys until the api returns back
             that the vault has been unsealed

    CLI Examples:

    .. code-block:: bash

        salt-run vault.unseal
    """
    for key in __opts__["vault"]["keys"]:
        ret = __utils__["vault.make_request"](
            "PUT", "v1/sys/unseal", data=json.dumps({"key": key})
        ).json()
        if ret["sealed"] is False:
            return True
    return False


def show_policies(minion_id):
    """
    Show the Vault policies that are applied to tokens for the given minion

    minion_id
        The minions id

    CLI Example:

    .. code-block:: bash

        salt-run vault.show_policies myminion
    """
    config = __opts__["vault"]
    return _get_policies(minion_id, config)


def _validate_signature(minion_id, signature, impersonated_by_master):
    """
    Validate that either minion with id minion_id, or the master, signed the
    request
    """
    pki_dir = __opts__["pki_dir"]
    if impersonated_by_master:
        public_key = "{}/master.pub".format(pki_dir)
    else:
        public_key = "{}/minions/{}".format(pki_dir, minion_id)

    log.trace("Validating signature for %s", minion_id)
    signature = base64.b64decode(signature)
    if not salt.crypt.verify_signature(public_key, minion_id, signature):
        raise salt.exceptions.AuthenticationError(
            "Could not validate token request from {}".format(minion_id)
        )
    log.trace("Signature ok")


def _get_policies(minion_id, config):
    """
    Get the policies that should be applied to a token for minion_id
    """
    _, grains, _ = salt.utils.minions.get_minion_data(minion_id, __opts__)
    policy_patterns = config.get(
        "policies", ["saltstack/minion/{minion}", "saltstack/minions"]
    )
    mappings = {"minion": minion_id, "grains": grains or {}}

    policies = []
    for pattern in policy_patterns:
        try:
            for expanded_pattern in _expand_pattern_lists(pattern, **mappings):
                policies.append(
                    expanded_pattern.format(**mappings).lower()  # Vault requirement
                )
        except KeyError:
            log.warning("Could not resolve policy pattern %s", pattern)

    log.debug("%s policies: %s", minion_id, policies)
    return policies


def _expand_pattern_lists(pattern, **mappings):
    """
    Expands the pattern for any list-valued mappings, such that for any list of
    length N in the mappings present in the pattern, N copies of the pattern are
    returned, each with an element of the list substituted.

    pattern:
        A pattern to expand, for example ``by-role/{grains[roles]}``

    mappings:
        A dictionary of variables that can be expanded into the pattern.

    Example: Given the pattern `` by-role/{grains[roles]}`` and the below grains

    .. code-block:: yaml

        grains:
            roles:
                - web
                - database

    This function will expand into two patterns,
    ``[by-role/web, by-role/database]``.

    Note that this method does not expand any non-list patterns.
    """
    expanded_patterns = []
    f = string.Formatter()

    # This function uses a string.Formatter to get all the formatting tokens from
    # the pattern, then recursively replaces tokens whose expanded value is a
    # list. For a list with N items, it will create N new pattern strings and
    # then continue with the next token. In practice this is expected to not be
    # very expensive, since patterns will typically involve a handful of lists at
    # most.

    for (_, field_name, _, _) in f.parse(pattern):
        if field_name is None:
            continue
        (value, _) = f.get_field(field_name, None, mappings)
        if isinstance(value, list):
            token = "{{{0}}}".format(field_name)
            expanded = [pattern.replace(token, str(elem)) for elem in value]
            for expanded_item in expanded:
                result = _expand_pattern_lists(expanded_item, **mappings)
                expanded_patterns += result
            return expanded_patterns
    return [pattern]


def _selftoken_expired():
    """
    Validate the current token exists and is still valid
    """
    try:
        verify = __opts__["vault"].get("verify", None)
        # Vault Enterprise requires a namespace
        namespace = __opts__["vault"].get("namespace")
        url = "{}/v1/auth/token/lookup-self".format(__opts__["vault"]["url"])
        if "token" not in __opts__["vault"]["auth"]:
            return True
        headers = {"X-Vault-Token": __opts__["vault"]["auth"]["token"]}
        # Add Vault namespace to headers if Vault Enterprise enabled
        if namespace is not None:
            headers["X-Vault-Namespace"] = namespace
        response = requests.get(url, headers=headers, verify=verify)
        if response.status_code != 200:
            return True
        return False
    except Exception as e:  # pylint: disable=broad-except
        raise salt.exceptions.CommandExecutionError(
            "Error while looking up self token : {}".format(str(e))
        )


def _get_token_create_url(config):
    """
    Create Vault url for token creation
    """
    role_name = config.get("role_name", None)
    auth_path = "/v1/auth/token/create"
    base_url = config["url"]
    return "/".join(x.strip("/") for x in (base_url, auth_path, role_name) if x)



"""
unittests for terraform roster
"""
import os.path

import salt.config
import salt.loader
from salt.roster import terraform
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class TerraformTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.roster.terraform
    """

    def setup_loader_modules(self):
        opts = salt.config.master_config(
            os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "master")
        )
        utils = salt.loader.utils(opts, whitelist=["roster_matcher"])
        return {terraform: {"__utils__": utils, "__opts__": {}}}

    def test_default_output(self):
        """
        Test the output of a fixture tfstate file which contains libvirt
        resources.
        """
        tfstate = os.path.join(
            os.path.dirname(__file__), "terraform.data", "terraform.tfstate"
        )
        pki_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "terraform.data")
        )

        with patch.dict(
            terraform.__opts__, {"roster_file": tfstate, "pki_dir": pki_dir}
        ):
            expected_result = {
                "db0": {
                    "host": "192.168.122.174",
                    "user": "root",
                    "passwd": "dbpw",
                    "tty": True,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
                "db1": {
                    "host": "192.168.122.190",
                    "user": "root",
                    "passwd": "dbpw",
                    "tty": True,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
                "web0": {
                    "host": "192.168.122.106",
                    "user": "root",
                    "passwd": "linux",
                    "timeout": 22,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
                "web1": {
                    "host": "192.168.122.235",
                    "user": "root",
                    "passwd": "linux",
                    "timeout": 22,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
            }

            ret = terraform.targets("*")
            self.assertDictEqual(expected_result, ret)

    def test_default_matching(self):
        """
        Test the output of a fixture tfstate file which contains libvirt
        resources using matching
        """
        tfstate = os.path.join(
            os.path.dirname(__file__), "terraform.data", "terraform.tfstate"
        )
        pki_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "terraform.data")
        )

        with patch.dict(
            terraform.__opts__, {"roster_file": tfstate, "pki_dir": pki_dir}
        ):
            expected_result = {
                "web0": {
                    "host": "192.168.122.106",
                    "user": "root",
                    "passwd": "linux",
                    "timeout": 22,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
                "web1": {
                    "host": "192.168.122.235",
                    "user": "root",
                    "passwd": "linux",
                    "timeout": 22,
                    "priv": os.path.join(pki_dir, "ssh", "salt-ssh.rsa"),
                },
            }

            ret = terraform.targets("*web*")
            self.assertDictEqual(expected_result, ret)



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.pw_group as pw_group
import salt.utils.platform
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


class PwGroupTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test for salt.module.pw_group
    """

    def setup_loader_modules(self):
        return {pw_group: {"grinfo": {}}}

    def test_add(self):
        """
        Tests to add the specified group
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(pw_group.__salt__, {"cmd.run_all": mock}):
            self.assertTrue(pw_group.add("a"))

    def test_delete(self):
        """
        Tests to remove the named group
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(pw_group.__salt__, {"cmd.run_all": mock}):
            self.assertTrue(pw_group.delete("a"))

    @skipIf(salt.utils.platform.is_windows(), "grp not available on Windows")
    def test_info(self):
        """
        Tests to return information about a group
        """
        self.assertDictEqual(pw_group.info("name"), {})

        mock = MagicMock(
            return_value={
                "gr_name": "A",
                "gr_passwd": "B",
                "gr_gid": 1,
                "gr_mem": ["C", "D"],
            }
        )
        with patch.dict(pw_group.grinfo, mock):
            self.assertDictEqual(pw_group.info("name"), {})

    @skipIf(salt.utils.platform.is_windows(), "grp not available on Windows")
    def test_getent(self):
        """
        Tests for return info on all groups
        """
        mock_getent = [{"passwd": "x", "gid": 0, "name": "root"}]
        with patch.dict(pw_group.__context__, {"group.getent": mock_getent}):
            self.assertDictContainsSubset(
                {"passwd": "x", "gid": 0, "name": "root"}, pw_group.getent()[0]
            )

        mock = MagicMock(return_value="A")
        with patch.object(pw_group, "info", mock):
            self.assertEqual(pw_group.getent(True)[0], "A")

    def test_chgid(self):
        """
        tests to change the gid for a named group
        """
        mock = MagicMock(return_value=1)
        with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
            self.assertTrue(pw_group.chgid("name", 1))

        mock = MagicMock(side_effect=[1, 0])
        with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
            mock = MagicMock(return_value=None)
            with patch.dict(pw_group.__salt__, {"cmd.run": mock}):
                self.assertTrue(pw_group.chgid("name", 0))

        mock = MagicMock(side_effect=[1, 1])
        with patch.dict(pw_group.__salt__, {"file.group_to_gid": mock}):
            mock = MagicMock(return_value=None)
            with patch.dict(pw_group.__salt__, {"cmd.run": mock}):
                self.assertFalse(pw_group.chgid("name", 0))



"""
unit tests for the localfs cache
"""
import errno
import shutil
import tempfile

import salt.cache.localfs as localfs
import salt.payload
import salt.utils.files
from salt.exceptions import SaltCacheError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class LocalFSTest(TestCase, LoaderModuleMockMixin):
    """
    Validate the functions in the localfs cache
    """

    def setup_loader_modules(self):
        return {localfs: {}}

    def _create_tmp_cache_file(self, tmp_dir):
        """
        Helper function that creates a temporary cache file using localfs.store. This
        is to used to create DRY unit tests for the localfs cache.
        """
        self.addCleanup(shutil.rmtree, tmp_dir)
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            localfs.store(bank="bank", key="key", data="payload data", cachedir=tmp_dir)

    # 'store' function tests: 5

    def test_handled_exception_cache_dir(self):
        """
        Tests that a SaltCacheError is raised when the base directory doesn't exist and
        cannot be created.
        """
        with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
            with patch("tempfile.mkstemp", MagicMock(side_effect=Exception)):
                self.assertRaises(
                    Exception, localfs.store, bank="", key="", data="", cachedir=""
                )

    def test_unhandled_exception_cache_dir(self):
        """
        Tests that a SaltCacheError is raised when the base directory doesn't exist and
        cannot be created.
        """
        with patch("os.makedirs", MagicMock(side_effect=OSError(1, ""))):
            self.assertRaises(
                SaltCacheError, localfs.store, bank="", key="", data="", cachedir=""
            )

    def test_store_close_mkstemp_file_handle(self):
        """
        Tests that the file descriptor that is opened by os.open during the mkstemp call
        in localfs.store is closed before calling salt.utils.files.fopen on the filename.

        This test mocks the call to mkstemp, but forces an OSError to be raised when the
        close() function is called on a file descriptor that doesn't exist.
        """
        with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
            with patch("tempfile.mkstemp", MagicMock(return_value=(12345, "foo"))):
                self.assertRaises(
                    OSError, localfs.store, bank="", key="", data="", cachedir=""
                )

    def test_store_error_writing_cache(self):
        """
        Tests that a SaltCacheError is raised when there is a problem writing to the
        cache file.
        """
        with patch("os.makedirs", MagicMock(side_effect=OSError(errno.EEXIST, ""))):
            with patch("tempfile.mkstemp", MagicMock(return_value=("one", "two"))):
                with patch("os.close", MagicMock(return_value=None)):
                    with patch(
                        "salt.utils.files.fopen", MagicMock(side_effect=IOError)
                    ):
                        self.assertRaises(
                            SaltCacheError,
                            localfs.store,
                            bank="",
                            key="",
                            data="",
                            cachedir="",
                        )

    def test_store_success(self):
        """
        Tests that the store function writes the data to the serializer for storage.
        """
        # Create a temporary cache dir
        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # Use the helper function to create the cache file using localfs.store()
        self._create_tmp_cache_file(tmp_dir)

        # Read in the contents of the key.p file and assert "payload data" was written
        with salt.utils.files.fopen(tmp_dir + "/bank/key.p", "rb") as fh_:
            for line in fh_:
                self.assertIn(b"payload data", line)

    # 'fetch' function tests: 3

    def test_fetch_return_when_cache_file_does_not_exist(self):
        """
        Tests that the fetch function returns an empty dic when the cache key file
        doesn't exist.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            self.assertEqual(localfs.fetch(bank="", key="", cachedir=""), {})

    def test_fetch_error_reading_cache(self):
        """
        Tests that a SaltCacheError is raised when there is a problem reading the cache
        file.
        """
        with patch("os.path.isfile", MagicMock(return_value=True)):
            with patch("salt.utils.files.fopen", MagicMock(side_effect=IOError)):
                self.assertRaises(
                    SaltCacheError, localfs.fetch, bank="", key="", cachedir=""
                )

    def test_fetch_success(self):
        """
        Tests that the fetch function is able to read the cache file and return its data.
        """
        # Create a temporary cache dir
        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # Use the helper function to create the cache file using localfs.store()
        self._create_tmp_cache_file(tmp_dir)

        # Now fetch the data from the new cache key file
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            self.assertIn(
                "payload data",
                localfs.fetch(bank="bank", key="key", cachedir=tmp_dir),
            )

    # 'updated' function tests: 3

    def test_updated_return_when_cache_file_does_not_exist(self):
        """
        Tests that the updated function returns None when the cache key file doesn't
        exist.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            self.assertIsNone(localfs.updated(bank="", key="", cachedir=""))

    def test_updated_error_when_reading_mtime(self):
        """
        Tests that a SaltCacheError is raised when there is a problem reading the mtime
        of the cache file.
        """
        with patch("os.path.isfile", MagicMock(return_value=True)):
            with patch("os.path.getmtime", MagicMock(side_effect=IOError)):
                self.assertRaises(
                    SaltCacheError, localfs.updated, bank="", key="", cachedir=""
                )

    def test_updated_success(self):
        """
        Test that the updated function returns the modification time of the cache file
        """
        # Create a temporary cache dir
        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # Use the helper function to create the cache file using localfs.store()
        self._create_tmp_cache_file(tmp_dir)

        with patch("os.path.join", MagicMock(return_value=tmp_dir + "/bank/key.p")):
            self.assertIsInstance(
                localfs.updated(bank="bank", key="key", cachedir=tmp_dir), int
            )

    # 'flush' function tests: 4

    def test_flush_key_is_none_and_no_target_dir(self):
        """
        Tests that the flush function returns False when no key is passed in and the
        target directory doesn't exist.
        """
        with patch("os.path.isdir", MagicMock(return_value=False)):
            self.assertFalse(localfs.flush(bank="", key=None, cachedir=""))

    def test_flush_key_provided_and_no_key_file_false(self):
        """
        Tests that the flush function returns False when a key file is provided but
        the target key file doesn't exist in the cache bank.
        """
        with patch("os.path.isfile", MagicMock(return_value=False)):
            self.assertFalse(localfs.flush(bank="", key="key", cachedir=""))

    def test_flush_success(self):
        """
        Tests that the flush function returns True when a key file is provided and
        the target key exists in the cache bank.
        """
        with patch("os.path.isfile", MagicMock(return_value=True)):
            # Create a temporary cache dir
            tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

            # Use the helper function to create the cache file using localfs.store()
            self._create_tmp_cache_file(tmp_dir)

            # Now test the return of the flush function
            with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
                self.assertTrue(localfs.flush(bank="bank", key="key", cachedir=tmp_dir))

    def test_flush_error_raised(self):
        """
        Tests that a SaltCacheError is raised when there is a problem removing the
        key file from the cache bank
        """
        with patch("os.path.isfile", MagicMock(return_value=True)):
            with patch("os.remove", MagicMock(side_effect=OSError)):
                self.assertRaises(
                    SaltCacheError,
                    localfs.flush,
                    bank="",
                    key="key",
                    cachedir="/var/cache/salt",
                )

    # 'list' function tests: 3

    def test_list_no_base_dir(self):
        """
        Tests that the ls function returns an empty list if the bank directory
        doesn't exist.
        """
        with patch("os.path.isdir", MagicMock(return_value=False)):
            self.assertEqual(localfs.list_(bank="", cachedir=""), [])

    def test_list_error_raised_no_bank_directory_access(self):
        """
        Tests that a SaltCacheError is raised when there is a problem accessing the
        cache bank directory.
        """
        with patch("os.path.isdir", MagicMock(return_value=True)):
            with patch("os.listdir", MagicMock(side_effect=OSError)):
                self.assertRaises(SaltCacheError, localfs.list_, bank="", cachedir="")

    def test_list_success(self):
        """
        Tests the return of the ls function containing bank entries.
        """
        # Create a temporary cache dir
        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # Use the helper function to create the cache file using localfs.store()
        self._create_tmp_cache_file(tmp_dir)

        # Now test the return of the ls function
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            self.assertEqual(localfs.list_(bank="bank", cachedir=tmp_dir), ["key"])

    # 'contains' function tests: 1

    def test_contains(self):
        """
        Test the return of the contains function when key=None and when a key
        is provided.
        """
        # Create a temporary cache dir
        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        # Use the helper function to create the cache file using localfs.store()
        self._create_tmp_cache_file(tmp_dir)

        # Now test the return of the contains function when key=None
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            self.assertTrue(localfs.contains(bank="bank", key=None, cachedir=tmp_dir))

        # Now test the return of the contains function when key='key'
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            self.assertTrue(localfs.contains(bank="bank", key="key", cachedir=tmp_dir))

    def test_mix_of_utf8_and_non_utf8_can_be_round_tripped(self):

        tmp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)

        data = {
            # Any unicode, which ideally is invalid ascii.
            "unicode": "áéí",
            # Any bytes so long as they're not valid utf-8
            "bytes": b"\xfe\x99\x00\xff",
        }
        bank = "bank"
        key = "key"

        self.addCleanup(shutil.rmtree, tmp_dir)
        with patch.dict(localfs.__opts__, {"cachedir": tmp_dir}):
            localfs.store(bank, key, data, tmp_dir)

            actual = localfs.fetch(bank, key, tmp_dir)

        self.assertEqual(data, actual)



"""
Test case for env sdb module
"""

import os
import textwrap

import pytest
import salt.utils.files
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS


@pytest.mark.windows_whitelisted
class EnvTestCase(ModuleCase, SaltReturnAssertsMixin):
    def setUp(self):
        self.state_name = "test_sdb_env"
        self.state_file_name = self.state_name + ".sls"
        self.state_file_set_var = os.path.join(
            RUNTIME_VARS.BASE_FILES, self.state_file_name
        )
        with salt.utils.files.fopen(self.state_file_set_var, "w") as wfh:
            wfh.write(
                textwrap.dedent(
                    """\
                set some env var:
                  cmd.run:
                    - name: echo {{ salt['sdb.set']('sdb://osenv/foo', 'bar') }}
                    - order: 1

                {% if salt['sdb.get']('sdb://osenv/foo') == 'bar' %}
                always-changes-and-succeeds:
                  test.succeed_with_changes:
                    - name: foo
                {% else %}
                always-changes-and-fails:
                  test.fail_with_changes:
                    - name: foo
                {% endif  %}
                """
                )
            )

    def tearDown(self):
        os.remove(self.state_file_set_var)

    @pytest.mark.slow_test
    def test_env_module_sets_key(self):
        state_key = "test_|-always-changes-and-succeeds_|-foo_|-succeed_with_changes"
        ret = self.run_function("state.sls", [self.state_name])
        self.assertTrue(ret[state_key]["result"])



"""
Test the ssh_known_hosts states
"""

import os
import shutil

import pytest
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS

GITHUB_FINGERPRINT = "9d:38:5b:83:a9:17:52:92:56:1a:5e:c4:d4:81:8e:0a:ca:51:a2:64:f1:74:20:11:2e:f8:8a:c3:a1:39:49:8f"
GITHUB_IP = "140.82.121.4"


@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
class SSHKnownHostsStateTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the ssh state
    """

    @classmethod
    def setUpClass(cls):
        cls.known_hosts = os.path.join(RUNTIME_VARS.TMP, "known_hosts")

    def tearDown(self):
        if os.path.isfile(self.known_hosts):
            os.remove(self.known_hosts)
        super().tearDown()

    @pytest.mark.slow_test
    def test_present(self):
        """
        ssh_known_hosts.present
        """
        kwargs = {
            "name": "github.com",
            "user": "root",
            "enc": "ssh-rsa",
            "fingerprint": GITHUB_FINGERPRINT,
            "config": self.known_hosts,
        }
        # test first
        ret = self.run_state("ssh_known_hosts.present", test=True, **kwargs)
        self.assertSaltNoneReturn(ret)

        # save once, new key appears
        ret = self.run_state("ssh_known_hosts.present", **kwargs)
        try:
            self.assertSaltTrueReturn(ret)
        except AssertionError:
            self.assertInSaltComment("Unable to receive remote host key", ret)
            self.skipTest("Unable to receive remote host key")

        self.assertSaltStateChangesEqual(
            ret, GITHUB_FINGERPRINT, keys=("new", 0, "fingerprint")
        )

        # save twice, no changes
        self.run_state("ssh_known_hosts.present", **kwargs)

        # test again, nothing is about to be changed
        ret = self.run_state("ssh_known_hosts.present", test=True, **kwargs)
        self.assertSaltTrueReturn(ret)

        # then add a record for IP address
        # pylint: disable=repeated-keyword
        ret = self.run_state("ssh_known_hosts.present", **dict(kwargs, name=GITHUB_IP))
        # pylint: enable=repeated-keyword
        try:
            self.assertSaltStateChangesEqual(
                ret, GITHUB_FINGERPRINT, keys=("new", 0, "fingerprint")
            )
        except AssertionError:
            self.assertInSaltComment("Unable to receive remote host key", ret)
            self.skipTest("Unable to receive remote host key")

        # record for every host must be available
        ret = self.run_function(
            "ssh.get_known_host_entries",
            ["root", "github.com"],
            config=self.known_hosts,
        )[0]
        try:
            self.assertNotIn(ret, ("", None))
        except AssertionError:
            raise AssertionError("Salt return '{}' is in ('', None).".format(ret))
        ret = self.run_function(
            "ssh.get_known_host_entries", ["root", GITHUB_IP], config=self.known_hosts
        )[0]
        try:
            self.assertNotIn(ret, ("", None, {}))
        except AssertionError:
            raise AssertionError(
                "Salt return '{}' is in ('', None,".format(ret) + " {})"
            )

    @pytest.mark.slow_test
    def test_present_fail(self):
        # save something wrong
        ret = self.run_state(
            "ssh_known_hosts.present",
            name="github.com",
            user="root",
            fingerprint="aa:bb:cc:dd",
            config=self.known_hosts,
        )
        self.assertSaltFalseReturn(ret)

    @pytest.mark.slow_test
    def test_absent(self):
        """
        ssh_known_hosts.absent
        """
        known_hosts = os.path.join(RUNTIME_VARS.FILES, "ssh", "known_hosts")
        shutil.copyfile(known_hosts, self.known_hosts)
        if not os.path.isfile(self.known_hosts):
            self.skipTest(
                "Unable to copy {} to {}".format(known_hosts, self.known_hosts)
            )

        kwargs = {"name": "github.com", "user": "root", "config": self.known_hosts}
        # test first
        ret = self.run_state("ssh_known_hosts.absent", test=True, **kwargs)
        self.assertSaltNoneReturn(ret)

        # remove once, the key is gone
        ret = self.run_state("ssh_known_hosts.absent", **kwargs)
        self.assertSaltStateChangesEqual(
            ret, GITHUB_FINGERPRINT, keys=("old", 0, "fingerprint")
        )

        # remove twice, nothing has changed
        ret = self.run_state("ssh_known_hosts.absent", **kwargs)
        self.assertSaltStateChangesEqual(ret, {})

        # test again
        ret = self.run_state("ssh_known_hosts.absent", test=True, **kwargs)
        self.assertSaltTrueReturn(ret)



import logging

log = logging.getLogger(__name__)


def test_get_schedule(schedule):
    """
    verify that the _get_schedule function works
    when remove_hidden is True and schedule data
    contains enabled key
    """

    job_name = "test_get_schedule"
    job = {
        "schedule": {
            "enabled": True,
            job_name: {"function": "test.ping", "seconds": 60},
        }
    }
    # Add the job to the scheduler
    schedule.opts.update(job)

    ret = schedule._get_schedule(remove_hidden=True)
    assert job["schedule"] == ret
